/*
g3d - calculate projection data in 3-dimensional fields 
*/

static char rcsid[] = "$Id: g3d.c,v 1.19 2008/03/07 18:55:50 carl Exp $";

#include <math.h>
#include <parse.h>
#include <stdlib.h>
#include <string.h>
#include "ct.h"

#define melp 20     /*max number of ellipsoids*/
#define ma 25       /*max number of azimuthal or z aperture blocks*/
#define md 1024      /*max number of detectors*/

/*start DSP*/
#define MAX(A,B) ((A)>(B) ? (A):(B))
#define MIN(A,B) ((A)>(B) ? (B):(A))
/*end DSP*/

#define DEBUG
#ifdef DEBUG
#define PDT(x)	parse_pdt(x);
#endif

static float rs=833.4;		/*source to iso distance [mm]*/
static float rd=1326.26;	/*source to detector distance [mm]*/
static float hs=2.6;		/*height of the source [mm]*/
static float ws=3.0;		/*width of source [mm]*/
static float hd=5.6;		/*height of the detector [mm]*/
static float zdet=0;		/*z-offset of detector row*/
static float sw=10.0;		/*slice width*/
static float wd=5.6;		/*width of detector [mm] [center-to-center]*/
static float da=100;		/*detector aperture [% of wd]*/
static float alpha=0.0;		/*first projection angle [degrees]*/
static int nrot=1;		/*# 360 rotations*/
static int nzs=1;		/*source-lets height [odd]*/
static int nas=1;		/*source-lets width [odd]*/
static int nzd=1;		/*detector-lets height*/
static int nad=1;		/*detector-lets width*/
static float slide=0;		/*z-slide [mm] per 360 rotation*/
static float zjump=0;		/*z jump after one rotation */
static int algo=CT_ANGLE;	/*0=parallel 1=fan-equispace 2=fan-equiangle */
static float target=90;		/*target angle (degrees)*/
static int twopi=0;		/*1=2pi parallel mode */
static int n=252;		/*number of symmetric detectors*/
static int nasym=0;		/*number of asymmetric detectors*/
static int ndeta;		/*total number of detectors (n-nasym)*/
static int k=360;		/*number of projections*/
static int eproj=0;		/*1=integreate exp[-proj]*/
static char *ifn="sim3.sf";	/*input file name*/
static char *ofn="proj";	/*output file name*/
static int print=1;		/*flag for printing program info*/
static int noise=0;		/*1=generate additive noise*/
static float skew=0;		/* DAS skew (percent of view angle) */
static float xo=0.0;		/*off-set parameters*/
static float yo=0.0;
static float zo=0.0;
static int var=0;		/*1=variable speed*/
static float vbreak=.1;		/*break point for variable speed*/
static char *sfn;		/*name of speed file for variable speed*/
static int mbct=0;		/*1= moving beam ct */
static int tstudy;		/*1=study table positioning errors */
static float tgain=1.0;		/* gain of table error function */
static float tfreq=0.5;		/* frequency of table error function */
static float tphase=0.;		/* phase of table error function (degrees)*/
static int speed=0;		/* speed trick */
static float pp=100;		/* penumbra percentage */
static float clamp=100;		/* % of detector usuable for mbhs */
static int debug;		/* 1=debug mode */
static float cl_cen;		/* center from command line */
static int cen_forced=0;	/* 1=center set in scanner= */
static int r_forced=0;		/* 1=r set in scanner= */
static char scanner[128];	/* arg from scanner= */
static float cl_r;		/* radius from command line */
static int quartdet=0;		/* switch quarter det offset */
static int pnview=0;		/* print every pnview views */
static long init_seed = 57891;	/* initial seed for random number gen */
static int additive=0;		/* 1=additive, 0=sorting */

static enum {			/* program mode */
	PROJ,		/* generate projections */
	IMAGE		/* generate image */
	} pmode;

static float x[3][melp],xi[3];		/*position of objects*/
static float d[3][melp],di[3];		/*length of ellipsoid axes*/

static float q[10][melp];		/*coefficients of ellipsoids*/
static float at[melp];			/*attenuation values*/
static int m;				/*number of ellipsoids*/
static char shapetype[melp];		/*type of shape*/
static float cx1[3][melp],cx1i[3];	/*position of endpoints for cylinder or cone*/
static float cx2[3][melp],cx2i[3];
static float norm[3][melp];		/*normal to plane of cyl or cone vector*/

					/* centers and normals to box */
static float cxy[6][melp],nxy[3][melp]; 
static float cxz[6][melp],nxz[3][melp]; 
static float cyz[6][melp],nyz[3][melp]; 

					/* plane displacement vectors */
static float dxy[2][melp];			
static float dxz[2][melp];			
static float dyz[2][melp];			
/*
	variables for image simulation
*/
static float offset=0;		/*image offset*/
static float pscale=1.0;	/*picture scale*/
static float xgrid=0.0;		/*x center of grid*/
static float ygrid=0.0;		/*y center of grid*/
static float zgrid=0.0;		/*z center of grid*/
static int nx=128;		/*size reconstruction grid - x*/
static int ny=128;		/*size reconstruction grid - y*/
static float xfov;		/* x - field of view */
static float yfov;		/* x - field of view */
static float pixel;		/* pixel size; override fov */
static int line=0;		/*1=line reconstruction*/
static float ipsi=0;		/* first rotation about z (degrees) */
static float itheta=0;		/* second rotation about x (degrees) */
static float iphi=0;		/* third rotation about z (degrees) */
static int ilet=1;		/* axial image lets */
static float px,py;		/* pixel sizes */

static void scanner_arg(char *op,char *arg);

struct PARSE_STRUCT parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input (shape) file"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"n",PARSE_INT,(PARSE_ARG)&n,"number of detectors"},
	{"nasym",PARSE_INT,(PARSE_ARG)&nasym,"number of asymmetric detectors"},
	{"k",PARSE_INT,(PARSE_ARG)&k,"number of views"},
	{"rs",PARSE_FLOAT,(PARSE_ARG)&rs,"source-center distance"},
	{"rd",PARSE_FLOAT,(PARSE_ARG)&rd,"source-detector distance"},
	{"hs",PARSE_FLOAT,(PARSE_ARG)&hs,"height of source"},
	{"hd",PARSE_FLOAT,(PARSE_ARG)&hd,"height of detector"},
	{"ws",PARSE_FLOAT,(PARSE_ARG)&ws,"width of source"},
	{"zdet",PARSE_FLOAT,(PARSE_ARG)&zdet,"z offset for detectors"},
	{"wd",PARSE_FLOAT,(PARSE_ARG)&wd,"width of detector"},
	{"sw",PARSE_FLOAT,(PARSE_ARG)&sw,"slice width"},
	{"xo",PARSE_FLOAT,(PARSE_ARG)&xo,"object displacement - x"},
	{"yo",PARSE_FLOAT,(PARSE_ARG)&yo,"object displacement - y"},
	{"zo",PARSE_FLOAT,(PARSE_ARG)&zo,"object displacement - z"},
	{"cen",PARSE_FLOAT,(PARSE_ARG)&cl_cen,"center detector"},
	{"r",PARSE_FLOAT,(PARSE_ARG)&cl_r,"scan radius"},
	{"sleth",PARSE_INT,(PARSE_ARG)&nzs,"source-lets height"},
	{"sletw",PARSE_INT,(PARSE_ARG)&nas,"source-lets width"},
	{"dleth",PARSE_INT,(PARSE_ARG)&nzd,"detector-lets height"},
	{"dletw",PARSE_INT,(PARSE_ARG)&nad,"detector-lets width"},
	{"pnview",PARSE_INT,(PARSE_ARG)&pnview,"print every pnview views"},
	{"seed",PARSE_INT,(PARSE_ARG)&init_seed,"init seed for random number generator"},
	{"seed",PARSE_SET,(PARSE_ARG)&noise},
	{"da",PARSE_FLOAT,(PARSE_ARG)&da,"active detector aperture (percent)"},
	{"slide",PARSE_FLOAT,(PARSE_ARG)&slide,"table travel in one rotation"},
	{"alpha",PARSE_FLOAT,(PARSE_ARG)&alpha,"first projection angle"},
	{"nrot",PARSE_INT,(PARSE_ARG)&nrot,"# 360 degree rotations"},
	{"zjump",PARSE_FLOAT,(PARSE_ARG)&zjump,"z displacement after rotation"},
	{"vbreak",PARSE_FLOAT,(PARSE_ARG)&vbreak,"variable speed parameter"},
	{"vbreak",PARSE_SET,(PARSE_ARG)&var},
	{"sf",PARSE_CHAR,(PARSE_ARG)&sfn,"file for saving displacements"},
	{"sf",PARSE_SET,(PARSE_ARG)&var},
	{"tgain",PARSE_FLOAT,(PARSE_ARG)&tgain,"table study parameter"},
	{"tgain",PARSE_SET,(PARSE_ARG)&tstudy},
	{"tfreq",PARSE_FLOAT,(PARSE_ARG)&tfreq,"table study parameter"},
	{"tfreq",PARSE_SET,(PARSE_ARG)&tstudy},
	{"tphase",PARSE_FLOAT,(PARSE_ARG)&tphase,"table study parameter"},
	{"tphase",PARSE_SET,(PARSE_ARG)&tstudy},
	{"pp",PARSE_FLOAT,(PARSE_ARG)&pp,"penumbra percentage"},
	{"clamp",PARSE_FLOAT,(PARSE_ARG)&clamp,"clamp distance for mbhs"},
	{"clamp",PARSE_SET,(PARSE_ARG)&mbct},
	{"target",PARSE_FLOAT,(PARSE_ARG)&target,"target angle (degrees)"},
	{"skew",PARSE_FLOAT,(PARSE_ARG)&skew,"DAS skew (percent)"},
	{"scanner",PARSE_PROC,(PARSE_ARG)scanner_arg,"sets scanner type"},

				/* options for simulated image */
	{"npic",PARSE_INT,(PARSE_ARG)&nx,"size of row/col of image"},
	{"npic",PARSE_INT,(PARSE_ARG)&ny},
	{"npic",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"nx",PARSE_INT,(PARSE_ARG)&nx,"x size of image"},
	{"nx",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"ny",PARSE_INT,(PARSE_ARG)&ny,"y size of image"},
	{"ny",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"xgrid",PARSE_FLOAT,(PARSE_ARG)&xgrid,"x grid center"},
	{"xgrid",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"ygrid",PARSE_FLOAT,(PARSE_ARG)&ygrid,"y grid center"},
	{"ygrid",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"zgrid",PARSE_FLOAT,(PARSE_ARG)&zgrid,"z grid center"},
	{"zgrid",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"pixel",PARSE_FLOAT,(PARSE_ARG)&pixel,"pixel size"},
	{"pixel",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"fov",PARSE_FLOAT,(PARSE_ARG)&xfov,"field of view"},
	{"fov",PARSE_FLOAT,(PARSE_ARG)&yfov},
	{"fov",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"xfov",PARSE_FLOAT,(PARSE_ARG)&xfov,"x field of view"},
	{"xfov",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"yfov",PARSE_FLOAT,(PARSE_ARG)&yfov,"y field of view"},
	{"yfov",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"scale",PARSE_FLOAT,(PARSE_ARG)&pscale,"image scale factor"},
	{"scale",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"offset",PARSE_FLOAT,(PARSE_ARG)&offset,"image offset"},
	{"offset",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"ipsi",PARSE_FLOAT,&ipsi,"first rotation about z (degrees)"},
	{"ipsi",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"itheta",PARSE_FLOAT,&itheta,"second rotation about x (degrees)"},
	{"itheta",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"iphi",PARSE_FLOAT,&iphi,"third rotation about z (degrees)"},
	{"iphi",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"ilet",PARSE_INT,&ilet,"axial lets for image"},
	{"ilet",PARSE_VALUE(IMAGE),(PARSE_ARG)&pmode},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'a',&algo,PARSE_FLAG_VALUE(CT_ANGLE),"equal angle fan beam projections"},
	{'A',&additive,PARSE_FLAG_SET,"additive attenuation"},
	{'s',&algo,PARSE_FLAG_VALUE(CT_SPACE),"equal space fan beam projections"},
	{'P',&algo,PARSE_FLAG_VALUE(CT_PAR),"parallel projections"},
	{'2',&twopi,PARSE_FLAG_SET,"2pi parallel projections"},
	{'d',&debug,PARSE_FLAG_SET,"turn on debug mode"},
	{'e',&eproj,PARSE_FLAG_SET,"integrate energy"},
	{'m',&mbct,PARSE_FLAG_SET,"moving beam helical"},
	{'n',&noise,PARSE_FLAG_SET,"add noise"},
	{'p',&print,PARSE_FLAG_CLR,"don't print status information"},
	{'S',&speed,PARSE_FLAG_SET,"use speed trick"},
	{'t',&tstudy,PARSE_FLAG_SET,"study table poistioning errors"},
	{'q',&quartdet,PARSE_FLAG_SET,"quarter detector offset"},
	{'v',&var,PARSE_FLAG_SET,"variable speed helical"},

				/* flags for simulated image */
	{'l',&line,PARSE_FLAG_SET,"line reconstruction"},
	{'l',(int *)&pmode,PARSE_FLAG_VALUE(IMAGE)},
	{'i',(int *)&pmode,PARSE_FLAG_VALUE(IMAGE),"image mode"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

/* start DSP */
typedef struct sorted_t {
  float t;            /*sorted t values*/
  int   t1;           /*is the sorted t value a t1 (entry) ?*/
  int   index;        /*index of shape in sorted list*/
} tsort_type;

typedef struct seg_at_t {
  float at;           /*attenuation of segment*/
  int index;          /*index of shape*/
} seg_at_type;

static tsort_type tsort[2*melp];       /*contains t values in sorted order*/
static seg_at_type seg_at[2*melp-1];   /*ith element is attenuation 
					between ith t and i+1th t*/
static int tsortlen;                   /*length of tsort*/

/*
	tsort_init - prep for tsorting
*/
static void
tsort_init()
{
	tsortlen = 0;
}

/*
	tsort_insert - insertion sort of t values 
*/
static void
tsort_insert(float t1,float t2,int j)
{
	int c1,c2;	/* loopers */

	if (t1 == t2) return;

				/* find where to place t1 */
	c1 = 0;
	if (tsortlen > 0){		
		while ((tsort[c1].t <= t1)&&(c1 < tsortlen)){
			c1++;
		}
	}
				/* shift entries to make space for t1 */
	for(c2 = tsortlen-1; c2 >= c1; c2--){
		tsort[c2+1].t = tsort[c2].t;
		tsort[c2+1].t1 = tsort[c2].t1;
		tsort[c2+1].index = tsort[c2].index;
	}

				/* insert t1 */
	tsort[c1].t = t1;
	tsort[c1].t1 = 1;
	tsort[c1].index = j;
	tsortlen++;

				/* find where to place t2 */
	while ((tsort[c1].t < t2)&&(c1 < tsortlen)){
		c1++;
	}

				/* shift entries to make space for t2 */
	for(c2 = tsortlen-1; c2 >= c1; c2--){
		tsort[c2+1].t = tsort[c2].t;
		tsort[c2+1].t1 = tsort[c2].t1;
		tsort[c2+1].index = tsort[c2].index;
	}

				/* insert t2 */
	tsort[c1].t = t2;
	tsort[c1].t1 = 0;
	tsort[c1].index = j;
	tsortlen++;
}

/*
	tsort_attenuate - assign attenuation values to list of t's
*/
static double 
tsort_attenuate()
{
	int c1,c2;	/* loopers */
	double spv;	/* return value */

	if(tsortlen == 0) return(0.0);

			/*sanity check*/
	if (((tsortlen%2) == 1)||(tsortlen > 2*m)) {
		parse_er("g3d: tsort_attenuate error: length of list of t's");
	}

			/*sanity check*/
	for(c1 = 1; c1 < tsortlen; c1++) {
		if(tsort[c1-1].t > tsort[c1].t)
		   parse_er("g3d: tsort_attenuate error: order of list of t's");
	}

		/*calculate spv for intersecting objects using stacking order*/
	for(c1 = 0; c1 < tsortlen-1; c1++){
		seg_at[c1].at = 0.0;
		seg_at[c1].index = -1;
	}

	for(c1 = 0; c1 < tsortlen-1; c1++){
		if (tsort[c1].t1){
			c2 = c1+1;
			do{
				if(tsort[c1].index > seg_at[c2-1].index){
					seg_at[c2-1].at = at[tsort[c1].index];
					seg_at[c2-1].index = tsort[c1].index;
				}
				c2++;
			}while (tsort[c2-1].index != tsort[c1].index);
		  }
	}

	spv = 0.0;
	for(c1 = 0; c1 < tsortlen-1; c1++){
		spv += seg_at[c1].at * (tsort[c1+1].t - tsort[c1].t);
	}

	return(spv);
}

/* end DSP */

/*
	setint/float - set or print an integer/float
*/
static void
setint(int *v,int val,char *name,int mode)
{
	char *np;

	if(mode == 0){
		*v = val;
	}else{
		np = parse_malloc(strlen(name)+2);
		strcpy(np,name);
		strcat(np,":");
		parse_pri(np,val);
	}
}

static void
setfloat(float *v,float val,char *name,int mode)
{
	char *np;

	if(mode == 0){
		*v = val;
	}else{
		np = parse_malloc(strlen(name)+2);
		strcpy(np,name);
		strcat(np,":");
		parse_prf(np,val);
	}
}

/*
	scanner_arg - process option "scanner"
*/
static void
scanner_arg(char *op,char *arg)
{
	int parg;
	char *qp;

	qp = strchr(arg,'?');
	if(qp == NULL){
		parg = 0;
	}else{
		parg = 1;
		*qp = 0;
	}

	if(!*arg){
		printf("scanners: a, c, d, prince, p5, eds, pki-argus, l3-argus\n");
		exit(0);
	}
				/* editor tag: Peep peep PEEP */

	if(!strcmp(arg,"a")){
		setint(&n,461,"n",parg); 	/* output file contains 384 */
		setint(&k,2880,"k",parg);
		setint(&nasym,77,"nasym",parg);
		setfloat(&cl_cen,152.75,"cen",parg);
		setfloat(&cl_r,229.148,"r",parg);
		setfloat(&da,82.6,"da",parg);
		setfloat(&hd,28,"hd",parg);
		setfloat(&hs,0.7,"hs",parg);
		setfloat(&rd,845.24,"rd",parg);
		setfloat(&rs,475.0,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,12,"target",parg);
		setfloat(&wd,1.844,"wd",parg);
		setfloat(&ws,1.9,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"c")){
		setint(&n,755,"n",parg);		/* output file contains 696 */
		setint(&k,960,"k",parg);
		setint(&nasym,59,"nasym",parg);
		setfloat(&cl_cen,377.25,"cen",parg);
		setfloat(&cl_r,241.753,"r",parg);
		setfloat(&da,81.7,"da",parg);
		setfloat(&hd,28,"hd",parg);
		setfloat(&hs,0.7,"hs",parg);
		setfloat(&rd,950.1,"rd",parg);
		setfloat(&rs,515.0,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,8,"target",parg);
		setfloat(&wd,1.229,"wd",parg);
		setfloat(&ws,1.2,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"d")){
		setint(&n,896,"n",parg);
		setint(&k,1440,"k",parg);
		setint(&nasym,0,"nasym",parg);
		setfloat(&cl_cen,447.25,"cen",parg);
		setfloat(&cl_r,249.998,"r",parg);
		setfloat(&da,78.0,"da",parg);
		setfloat(&hd,1.766990,"hd",parg);		/* for 1-mm slice */
		setfloat(&hs,1.0,"hs",parg);
		setfloat(&rd,910.0,"rd",parg);
		setfloat(&rs,515.0,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,10,"target",parg);
		setfloat(&wd, 1.02956,"wd",parg);
		setfloat(&ws,1.2,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"prince")){ 	/* 8 x 1.25 version of D */
		setint(&n,896,"n",parg);
		setint(&k,1440,"k",parg);
		setint(&nasym,0,"nasym",parg);
		setfloat(&cl_cen,447.25,"cen",parg);
		setfloat(&cl_r,249.998,"r",parg);
		setfloat(&da,78.0,"da",parg);
		setfloat(&hd,1.766990*1.25,"hd",parg);		
		setfloat(&hs,1.0,"hs",parg);
		setfloat(&rd,910.0,"rd",parg);
		setfloat(&rs,515.0,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,10,"target",parg);
		setfloat(&wd, 1.02956,"wd",parg);
		setfloat(&ws,1.2,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"p5")){
		setint(&n,479,"n",parg); 	/* output=414 */
		setint(&k,1440,"k",parg);
		setint(&nasym,65,"nasym",parg);
		setfloat(&cl_cen,173.25,"cen",parg);
		setfloat(&cl_r,411,"r",parg);
		setfloat(&da,81.0,"da",parg);
		setfloat(&hd,19.6,"hd",parg);
		setfloat(&hs,0.7,"hs",parg);
		setfloat(&rd,845.0,"rd",parg);
		setfloat(&rs,495.0,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,10,"target",parg);
		setfloat(&wd,1.552,"wd",parg);
		setfloat(&ws,1.7,"ws",parg);

		r_forced = 1;
		cen_forced = 1;
	}else if(!strcmp(arg,"eds")){
		setint(&n,252,"n",parg);
		setint(&k,360,"k",parg);
		setint(&nasym,0,"nasym",parg);
		setfloat(&cl_cen,125.5,"cen",parg);
		setfloat(&cl_r,422.764,"r",parg);
		setfloat(&da,86.0,"da",parg);
		setfloat(&hd,4.816,"hd",parg); 	/* 5.6*0.86=4.816 mm */
		setfloat(&hs,2.6,"hs",parg);
		setfloat(&rd,1326.26,"rd",parg);
		setfloat(&rs,833.4,"rs",parg);
		setfloat(&slide,98.786667,"slide",parg);
		setfloat(&target,20,"target",parg);
		setfloat(&wd,5.6,"wd",parg);
		setfloat(&ws,3.0,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"pki-argus")){
		setint(&n,823,"n",parg);
		setint(&k,1440,"k",parg);
		setint(&nasym,0,"nasym",parg);
		setfloat(&cl_cen,410.75,"cen",parg);
		setfloat(&cl_r,450,"r",parg);
		setfloat(&da,87.9,"da",parg);
		setfloat(&hd,9.00,"hd",parg);
		setfloat(&hs,0.4,"hs",parg);
		setfloat(&rd,1295.60,"rd",parg);
		setfloat(&rs,740.96,"rs",parg);
		setfloat(&slide,0,"slide",parg);
		setfloat(&target,10,"target",parg);
		setfloat(&wd,2.06,"wd",parg);
		setfloat(&ws,1.5,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else if(!strcmp(arg,"l3-argus")){
		setint(&n,506,"n",parg);
		setint(&k,1440,"k",parg);
		setint(&nasym,0,"nasym",parg);
		setfloat(&cl_cen,253.25,"cen",parg); 	/* for QDO */
		setfloat(&cl_r,425.5,"r",parg); 		
		setfloat(&da,83.7,"da",parg);
		setfloat(&hd,5.079,"wd",parg);
		setfloat(&hs,0.4,"hs",parg);
		setfloat(&rd,1297.13,"rd",parg);
		setfloat(&rs,740.96,"rs",parg);
		setfloat(&slide,19.1522,"slide",parg);
		setfloat(&target,10,"target",parg);
		setfloat(&wd,3.124,"wd",parg);
		setfloat(&ws,1.5,"ws",parg);

		cen_forced = 1;
		r_forced = 1;
	}else{
		printf("g3d: illegal scanner: %s\n",arg);
		exit(1);
	}
	strncpy(scanner,arg,128);
	if(parg == 1)exit(0);
}


/*
	gen_image - generate simulated image (code based on sim.c)
*/
static void
gen_image(void)
{
	int i,j,l,zl;		/* counters and loopers */
	float *pic,*ppic;	/* output line as da-file */
	short *mi;		/* output line as mi-file */
	short *mipic;		/* complete picture */
	int ll;			/* number of rows */
	float pi;		/* good time of year tomato pie */
	float xs,ys;		/* starting locations for rows */
	float p;		/* projection value */
	float xf,yf,zf;		/* spatial coordinate in stationary (fixed) frame */
	float plane1,plane2;	/* dot products with bounding planes */
	int in;			/* =1 in object */
	float xp,yp,zp;		/* spatial coordinate in image plane */
	float convert;		/* degrees to radians */

	float cpsi,spsi;	/* first rotation about x */
	float stheta,ctheta;	/* second rotation about x */
	float cphi,sphi;	/* third rotation about z */
	float b00,b01,b02,b10,b11,b12,b20,b21,b22; /* rotation matrix */
	float pxy1,pxy2;	/* dot products with rotated xy planes */
	float pxz1,pxz2;	/* dot products with rotated xz planes */
	float pyz1,pyz2;	/* dot products with rotated yz planes */
/*
	set up general parameters
*/
        pi = M_PI;			/*this time of year it is frozen*/
        pscale *=  10240. / .188;	/*make .188=1024*/
        convert = pi / 180.0;
        xs = -(nx-1)/2.0 * px + xgrid;
	if(!line){
		ys = (ny-1)/2.0 * py + ygrid;
		ll = ny;                       /*# of lines to loop*/
		mi = parse_malloc_short(nx);
		mipic = parse_malloc_short(nx*ny);
	}else{
		ys = ygrid;                    
		ll = 1;
	}

	pic = parse_malloc_float(nx);
	ppic = parse_malloc_float(nx);
	if(ilet <= 0) parse_er("g3d: bad ilet");
/*
	set up rotation matrix
*/
	ipsi *= convert;
	iphi *= convert;
	itheta *= convert;
	
	cpsi = cos(ipsi);
	spsi = sin(ipsi);
	cphi = cos(iphi);
	sphi = sin(iphi);
	ctheta = cos(itheta);
	stheta = sin(itheta);

	b00 = cphi*cpsi - sphi*ctheta*spsi;
	b01 = cphi*spsi + sphi*ctheta*cpsi;
	b02 = sphi * stheta;
	b10 = -sphi * cpsi - cphi * ctheta * spsi;
	b11 = -sphi * spsi + cphi * ctheta * cpsi;
	b12 = cphi * stheta;
	b20 = stheta * spsi;
	b21 = -stheta * cpsi;
	b22 = ctheta;
/*
	generate picture by looping through it line by line
*/
	for(l=0;l<ll;l++){			/* loop lines */
	  for(j=0;j<nx;j++) pic[j] = 0;
	  yp = ys;

	  for(zl=0;zl<ilet;zl++){
	    zp = zgrid + -sw/2 + sw/ilet*(zl+.5);

	    for(j=0;j<nx;j++) ppic[j] = 0;

	    for(i=0;i<m;i++){		/* loop shapes */
	    	xp = xs;
	    	for(j=0;j<nx;j++){	/*loop though pixels*/
	    		xf = xp*b00 + yp*b01 + zp*b02;
	    		yf = xp*b10 + yp*b11 + zp*b12;
	    		zf = xp*b20 + yp*b21 + zp*b22;

	    		if(shapetype[i] == 'B'){
				pxy1 =   nxy[0][i]*(xf-cxy[0][i])+
					 nxy[1][i]*(yf-cxy[1][i])+
					 nxy[2][i]*(zf-cxy[2][i]);

				pxy2 =   nxy[0][i]*(xf-cxy[3][i])+
					 nxy[1][i]*(yf-cxy[4][i])+
					 nxy[2][i]*(zf-cxy[5][i]);

				pxz1 =   nxz[0][i]*(xf-cxz[0][i])+
					 nxz[1][i]*(yf-cxz[1][i])+
					 nxz[2][i]*(zf-cxz[2][i]);

				pxz2 =   nxz[0][i]*(xf-cxz[3][i])+
					 nxz[1][i]*(yf-cxz[4][i])+
					 nxz[2][i]*(zf-cxz[5][i]);

				pyz1 =   nyz[0][i]*(xf-cyz[0][i])+
					 nyz[1][i]*(yf-cyz[1][i])+
					 nyz[2][i]*(zf-cyz[2][i]);

				pyz2 =   nyz[0][i]*(xf-cyz[3][i])+
					 nyz[1][i]*(yf-cyz[4][i])+
					 nyz[2][i]*(zf-cyz[5][i]);

				
				if(pxy1*pxy2 >0 || pxz1*pxz2 >0 || pyz1*pyz2 >0){
					in = 0;
	    			}else{
	    			   	in = 1;
	    				p = pscale * at[i];
	    			}
	    		}else{
	    			if(0 > q[0][i]*xf*xf + q[1][i]*yf*yf + q[2][i]*zf*zf + 
	    			  2*q[3][i]*xf*yf + 2*q[4][i]*xf*zf + 2*q[5][i]*yf*zf + 
	    			  2*q[7][i]*xf + 2*q[8][i]*yf + 2*q[9][i]*zf + q[6][i]){
	    				in = 1;
	    				p = pscale * at[i];
	    			}else{
	    				in = 0;
	    			}

	    			if(shapetype[i] == 'C' || shapetype[i] == 'Y'){
	    				plane1 = norm[0][i]*(xf-cx1[0][i])+
	    					 norm[1][i]*(yf-cx1[1][i])+
	    					 norm[2][i]*(zf-cx1[2][i]);

	    				plane2 = norm[0][i]*(xf-cx2[0][i])+
	    					 norm[1][i]*(yf-cx2[1][i])+
	    					 norm[2][i]*(zf-cx2[2][i]);
	    				
	    				if(plane1*plane2 >0) in = 0;
	    			}
	    		}
	    		if(in){
				if(additive){
					ppic[j] += p;
				}else{
					ppic[j] = p;
				}
	    		}
	    		xp += px;
	    	}
	    }
	    for(j=0;j<nx;j++) pic[j] += ppic[j];
	  }

	  for(j=0;j<nx;j++) pic[j] = pic[j] / ilet + offset;

	  if(!line){
	  	for(j=0;j<nx;j++){
	  		mi[j] = pic[j];
	  		if(mi[j] < 0) mi[j] = 0.0;
	  		mipic[l*nx+j] = mi[j];
	  	}
	  }
	  ys -= py;
	}
/*
	write output files
*/
	if(line){
		parse_wda(pic,nx,1,ofn);
	}else{
		parse_wmi(ofn,mipic,nx,ny);
	}
/*	
	clean up and return
*/
	return;
}

int
main(int argc,char **argv)
{
        float r;			/*radius of the scan circle*/
        float psi[melp];		/*rotation angle about"z"*/
        float theta[melp];		/*rotation angle about"x1"*/
        float phi[melp];		/*rotation angle about"z2"*/
        float sind[md];			/*sin of each detector angle*/
        float cosd[md];			/*cos of each detector angle*/
        float sins[md];			/*sin of each source angle*/
        float coss[md];			/*cos of each source angle*/
        float adet[md];			/*angle of each detector*/
	float aproj[md],*proj;		/*resultant projections */
        float a[3][3],b[3][3],c[3][3]; 	/* rotation matricies*/
        float csa[ma*md];		/*azimuthal & z displacements */
        float ssa[ma*md];
        float cda[ma];
        float sda[ma];
        float zdiss[ma];
        float zdisd[ma];
	float dzw[ma];		/* detector z weight */
	float dzwp;
	char sbuf[PARSE_MAX_LINE];
	float cen;		/* center detector */
	float wdiso;		/* width of detector at iso-center */
	float wdziso;		/* width of detector in z at iso-center */
	float scale;		/* scale for output header */
	float cena;		/* center detector for asymmetric fan */
	float conea;		/* cone angle */

	float pi,psii,thetai,phii,ati;
	int i,j,l,ii,nblk,iblk;
        double b1j,b2j,b3j,b4j,b5j,b6j,b7j,b8j,b9j,b10j;
        double a1,a2,a3;
	float convert,g1,g2,g3,g4,g5,g6,g7,x1,x2,x3,ths,dths,helixz;
	float helixzp;
	float cths,sths,psum,thsp;
	float rdp,sclin,sclog;
	float dgamma,t,dt,dz,z,ggamma;
	float wdp,cpsi,spsi,cphi,sphi,ctheta,stheta;
	float nprod;
	double xs,ys,xs2,hs2;
	float cb,sb,xd,yd;
	double ax1,ax2,ax12,ax22;
	double spv,ax3,ax32,ys2,ar2;

	int istate,i1,i2,isa,isz,ida,idz;
	float *temp;
	FILE *iu,*ou;
	HU hu;			/*header for input and output files*/
	float ls,xb,xc,fwzm,fwfm,as,bs,cs;	/* focal spot weight info */
	int isap;		/* related to isa */
        float wsa[ma*md];		/* source-let weights */
/*start DSP*/
	char ch;
	float cr1[melp],cr1i;           /*radius of cylinders*/
	float cr2[melp],cr2i;
	float tc1,tc2,tp1,tp2,t1,t2;    /*t in parametric line*/
	float normB;                    /*plane normal dot product with line*/
	float plane1,plane2;            /*dot products*/
	int   tp1in,tp2in;              /*is tp1 or tp2 hitting the cyl/cone*/
/*end DSP*/

	float delt,dotb,tnx,tfx,tny,tfy,tnz,tfz;
	float rb[3];			/* ray's initial vector position */
	float ra[3];			/* ray's initial vector direction */
	float sqrt_huge;		/* for box tracing */
/*
	parse command line
*/
	parse_all_basic(".g3d","G3DARGS",argc,argv);
/*
	other initialization 
*/
        pi = M_PI;
	sqrt_huge = sqrt(HUGE_VAL);	/* for box tracing */
        if(noise) srand48(init_seed);
/*
	open, read, close ellpsoid file
*/
        ifn = parse_suffix_malloc(ifn,"sf");
	iu = parse_fopen(ifn,"r");
        m = 0;
	while(!parse_urline(iu,sbuf)){
/*start DSP*/
	        /* any line not starting with B, E, C or Y is considered a comment */
		if ( (*sbuf != 'E') &&
		     (*sbuf != 'C') &&
		     (*sbuf != 'Y') && 
		     (*sbuf != 'B') &&
		     (*sbuf != '#'))
			parse_err("g3d: syntax error in: ",ifn);

		if (*sbuf == '#'){ /* comment */
			continue;
		}

		if (*sbuf == 'B'){ /* box */
			if(11 != sscanf(sbuf,"%c%f%f%f%f%f%f%f%f%f%f",
			  &ch,xi,di,xi+1,di+1,xi+2,di+2,&psii,&thetai,&phii,&ati)){
				parse_err("g3d: syntax error (B) in: ",ifn);
			}
		} 
		if (*sbuf == 'E') /* ellipsoid */
		  if(11 != sscanf(sbuf,"%c%f%f%f%f%f%f%f%f%f%f",
		  &ch,xi,di,xi+1,di+1,xi+2,di+2,&psii,&thetai,&phii,&ati)){
		  	parse_err("syntax error (E) in: ",ifn);
		      }
		if (*sbuf == 'C') /* cone */
	 	  if(10 != sscanf(sbuf,"%c%f%f%f%f%f%f%f%f%f",
		  &ch,cx1i,cx1i+1,cx1i+2,cx2i,cx2i+1,cx2i+2,&cr1i,&cr2i,&ati)) { 
		        parse_err("syntax error (C) in: ",ifn);
		      }
		if (*sbuf == 'Y') /* cylinder */
		  if (9 != sscanf(sbuf,"%c%f%f%f%f%f%f%f%f",
		  &ch,cx1i,cx1i+1,cx1i+2,cx2i,cx2i+1,cx2i+2,&cr1i,&ati)) { 
		  	parse_err("syntax error (Y) in: ",ifn);
		      }
		if ((*sbuf == 'C')||(*sbuf == 'Y')) {
		  /* setup normal vector and rotations */
		  norm[0][m] = cx1i[0]-cx2i[0];
		  norm[1][m] = cx1i[1]-cx2i[1];
		  norm[2][m] = cx1i[2]-cx2i[2];
		  if (norm[0][m] < 0) {  /* make normal point in +x direction */
		    norm[0][m] *= -1;
		    norm[1][m] *= -1;
		    norm[2][m] *= -1;
		  }
		  phii = acos(MAX(-1.0,MIN(1.0,norm[0][m]/sqrt(norm[0][m]*norm[0][m]+
					      norm[1][m]*norm[1][m]+norm[2][m]*norm[2][m]))));
		  if ((norm[1][m] == 0.0)&&(norm[2][m] == 0.0)) /* normal in x direction */
		    thetai = 0.0; /* its actually indeterminate */
		  else
		    thetai = acos(MAX(-1.0,MIN(1.0,norm[1][m]/sqrt(norm[1][m]*norm[1][m]+
						  norm[2][m]*norm[2][m]))));
		  if (norm[2][m] < 0)  /* theta has a 360 range, this selects the right theta */
		    thetai *= -1;
		  phii *= 180.0 / pi;
		  thetai *= 180.0 / pi;
		  psii = 0.0;
		  /* setup translation and radii */
		  if (*sbuf == 'Y') {
		    xi[0] = cx1i[0];
		    xi[1] = cx1i[1];
		    xi[2] = cx1i[2];
		    /* sscanf("+Infinity","%f",&(di[0])); */
 		    /* di[0] never gets looked at.  it is trapped later */
		    di[0] =  -1;
		    di[1] = di[2] = cr1i;
		  }
		  if (*sbuf == 'C') {
		    xi[0] = (cr2i*cx1i[0]-cr1i*cx2i[0])/(cr2i-cr1i);
		    xi[1] = (cr2i*cx1i[1]-cr1i*cx2i[1])/(cr2i-cr1i);
		    xi[2] = (cr2i*cx1i[2]-cr1i*cx2i[2])/(cr2i-cr1i);
		    di[0] = 1.0;  /* sign gets changed later on */
		    di[1] = di[2] = cr1i/sqrt((cx1i[0]-xi[0])*(cx1i[0]-xi[0])+(cx1i[1]-xi[1])*(cx1i[1]-xi[1])+
					      (cx1i[2]-xi[2])*(cx1i[2]-xi[2]));
		  }
		  /* translate bounding planes by xo,yo,zo */
		  cx1[0][m] = cx1i[0]+xo;  cx2[0][m] = cx2i[0]+xo;
		  cx1[1][m] = cx1i[1]+yo;  cx2[1][m] = cx2i[1]+yo;
		  cx1[2][m] = cx1i[2]+zo;  cx2[2][m] = cx2i[2]+zo;
		  cr1[m] = cr1i;
		  cr2[m] = cr2i;
		}

                m++;
                parse_pcheck(m,1,melp,"g3d: too many objects");
		shapetype[m-1] = ch;
                at[m-1] = ati;
/*end DSP*/
		for(j=0;j<3;j++){
			x[j][m-1] = xi[j];
			d[j][m-1] = di[j];
		}
		psi[m-1] = psii;
		theta[m-1] = thetai;
		phi[m-1] = phii;
        }
	parse_fclose(iu);
/*
	check parameter validity
*/
        parse_pcheck(n,1,md,"bad detector count");
        parse_pcheck(nzs,1,ma,"bad sleth");
        parse_pcheck(nas,1,ma,"bad sletw");
        parse_pcheck(nzd,1,ma,"bad dleth");
        parse_pcheck(nad,1,ma,"bad dletw");
        if(target <= 0 || target > 90) parse_er("g3d: bad target");
	if(parse_opexist("sw") && parse_opexist("hd"))
		parse_er("g3d: can't set both 'hd' and 'sw'");

	if(m>1 && speed) 
		parse_er("g3d: can't use speed trick & multiple objects");

	if(m == 1) additive = 1;
/*
	set up geometries
*/
        convert = pi / 180.0;

	if(parse_opexist("cen") || cen_forced == 1){
		cen = cl_cen;
	}else{
		if(quartdet)cen = n / 2 - 1 + .75;
		else cen = n / 2 - 1 + .5;
	}
	cena = cen - nasym;

	if(parse_opexist("sw")) hd = sw * rd / rs;
	wdziso = hd * rs / rd;
	if(!parse_opexist("sw")) sw = wdziso;

	wdiso = wd * rs / rd;

	if(algo == CT_PAR){
		r = n / 2.0 * wdiso;
		dths = pi / k;	
		scale = M_PI / k;	
		if(twopi){
			dths *= 2.;
		}
	}else{
		wdiso = wd * rs / rd;
		dgamma = atan(wdiso/rs);

		if(algo == CT_ANGLE){
			ggamma = dgamma * n / 2.0;
		}else{
			ggamma = atan(wdiso * n / 2.0/rs);
		}
		r = rs * sin(ggamma);
		dths = (2.0 * pi) / k;	
		scale = M_PI / k;	/* see NOTES for more info */
	}

	if(r_forced == 1) r = cl_r;
	if(parse_opexist("r")) r = cl_r;

	conea = atan(zdet/rd);

	if(pmode == IMAGE){
		if(!parse_opexist("fov") && 
		   !parse_opexist("xfov") && 
		   !parse_opexist("yfov")){
			xfov = yfov = 2.0 * r;
		}

		if(parse_opexist("pixel")){
			xfov = nx * pixel;
			yfov = ny * pixel;
		}
		px = xfov / nx;
		py = yfov / ny;
	}
/*
	open output file and write header and fix other files for projection mode;
*/
        if(pmode == PROJ){
		ofn = parse_suffix_malloc(ofn,"da");
		ou = parse_fopen(ofn,"w");

		parse_zbuf(hu.heads,256);

		ndeta = n - nasym;

		hu.heads[CT_N] = ndeta;		/* number of detectors */
		hu.heads[CT_K] = k*nrot;	/* number of views */
		hu.heads[CT_ROT] = nrot;	/* # rotations */
		hu.heads[CT_MODE] = 7;		/* pass-through mode */
		hu.heads[CT_ALGO] = algo;	/* pass-through mode */

		hu.headf[CT_R] = r;		/* scan radius */
		hu.headf[CT_RSC] = rs;		/* source to center distance */
		hu.headf[CT_ALPHA] = alpha*convert; /* starting angle */
		hu.headf[CT_DBETA] = dths;	/* angle between projections */
		hu.headf[CT_SCALE] = scale;	/* scale factor */
		hu.headf[CT_SLIDE] = slide;	/* delta_z per 2 pi */
		hu.headf[CT_WDISO] = wdiso;	/* width of detector at iso */
		hu.headf[CT_VSTART] = 0.;	/* starting view */
		hu.headf[CT_CONEA] = conea;	/* cone angle */
		hu.headf[CT_WZISO] = wdziso;	/* width in z at iso */

		hu.headf[CT_CEN] = cena;	

		ct_write_header(ou,&hu);

		nblk = BLOCK_FLOAT(ndeta);
		iblk = 1;

		temp = parse_malloc_float(ndeta); /* temp vector for output */	

		if(parse_opexist("sf")) sfn = parse_suffix_malloc(sfn,"da");
	}
/*
	initialization for image mode
*/
	if(pmode == IMAGE){
		ofn = parse_suffix_malloc(ofn,line? "da" : "mi");
	}
/*
	print program status
*/
        if(print){
                parse_wtext("******************** g3d ********************");
                parse_pdt("---start time");
		parse_rcsid_print(rcsid);
                if(parse_opexist("scanner")) parse_prs("scanner:",scanner);
                parse_prs("output file:",ofn);
                parse_prs("input file :",ifn);

                if(pmode == PROJ){
			parse_wtext("mode: projection generation");
			parse_pri("n =",n);
			parse_pri("nasym =",nasym);
			parse_pri("k (per rot) =",k);
			parse_pri("nrot =",nrot);
			parse_prf("rs =",rs);
			parse_prf("rd =",rd);
			parse_prf("hs =",hs);
			parse_prf("ws =",ws);
			parse_prf("target (degree) =",target);
			parse_prf("DAS skew (percent) =",skew);
			parse_prt("calculate hd from sw",parse_opexist("sw"));
			parse_prf("sw =",sw);
			parse_prf("hd =",hd);
			parse_prf("wd =",wd);
			parse_prf("wd (iso) =",wdiso);
			parse_prf("zdet =",zdet);
			parse_prf("da (percent) =",da);
			parse_pri("sleth =",nzs);
			parse_pri("sletw =",nas);
			parse_pri("dleth =",nzd);
			parse_pri("dletw =",nad);
			parse_prf("r =",r);
			parse_prf("slide =",slide);
			parse_prf("zjump =",zjump);
			parse_prf("alpha =",alpha);
			parse_prt("exp(-proj)",eproj);
			parse_prt("debug",debug);
			parse_prt("quartdet",quartdet);
			parse_prf("center detector =",cen);
			parse_prf("cone angle =",conea*180/pi);
			parse_prf("wdziso =",wdziso);
			if(nasym>0)parse_prf("asymm center detector(output) =",cena);
			if(algo == CT_PAR){
				parse_wtext("parallel projections");
				parse_prt("2pi projections",twopi);
				parse_prf("dbeta =",dths);
			}else{
				parse_prf("fan angle (degrees) =",ggamma*180/pi);
				parse_prf("dgamma (degrees) =",dgamma*180/pi);
				parse_prf("dbeta =",dths*180/pi);
				if(algo == CT_SPACE){
					parse_wtext("fanbeam projections: equispace");
				}else{
					parse_wtext("fanbeam projections: equiangle");
				}
			}
			if(mbct){
				parse_prt("moving beam ct",mbct);
				parse_prf("detector clamp (%) =",clamp);
			}
			parse_prt("speed trick",speed);
			parse_prf("pp (penumbra percentage)=",pp);

			parse_prt("add noise",noise);
			if(var){
				parse_wtext("variable speed mode");
				parse_prf("vbreak =",vbreak);
				if(parse_opexist("sf")) parse_prs("speed file:",sfn);
			}
			if(tstudy){
				parse_wtext("study table positioning errors");
				parse_prf("tgain =",tgain);
				parse_prf("tfreq =",tfreq);
				parse_prf("tphase =",tphase);
			}
			parse_prs("attenuation:",additive? "additive":"sorting");
		}

		if(pmode == IMAGE){
			parse_wtext("mode: image generation");
			parse_pri("image offset:",offset);
			parse_prf("image scale:",pscale);
			parse_prf("xgrid =",xgrid);
			parse_prf("ygrid =",ygrid);
			parse_prf("zgrid =",zgrid);
			parse_pri("nx =",nx);
			parse_pri("ny =",ny);
			parse_prf("xfov =",xfov);
			parse_prf("yfov =",yfov);
			parse_prf("pixel-x =",px);
			parse_prf("pixel-y =",py);
			parse_prt("line",line);
			parse_prf("ipsi (degrees) =",ipsi);
			parse_prf("itheta (degrees) =",itheta);
			parse_prf("iphi (degrees) =",iphi);
			parse_pri("ilet =",ilet);
			parse_prf("sw =",sw);
		}
                parse_pri("m =",m);
                parse_prf("xo =",xo);
                parse_prf("yo =",yo);
                parse_prf("zo =",zo);
                parse_wtext("\nShape input data");

		for(i=0;i<m;i++){
			if(shapetype[i] == 'E'){
				printf("Object %i: - Ellipsoid\n",i+1);
				printf("x,  y,  z  = %f %f %f\n",x[0][i],x[1][i],x[2][i]);
				printf("ax, ay, az = %f %f %f\n",d[0][i],d[1][i],d[2][i]);
				printf("psi, theta, phi, mu = %f %f %f %f\n",psi[i],theta[i],phi[i],at[i]);
			}
			if(shapetype[i] == 'Y'){
				printf("Object: %i - Cylinder\n",i+1); 
				printf("x0, y0, z0 = %f %f %f\n",cx1[0][i],cx1[1][i],cx1[2][i]);
				printf("x1, y1, z1 = %f %f %f\n",cx2[0][i],cx2[1][i],cx2[2][i]);
				printf("r0, mu = %f %f\n",cr1[i],at[i]);
			}
			if(shapetype[i] == 'C'){
				printf("Object: %i - Cone\n",i+1); 
				printf("x0, y0, z0 = %f %f %f\n",cx1[0][i],cx1[1][i],cx1[2][i]);
				printf("x1, y1, z1 = %f %f %f\n",cx2[0][i],cx2[1][i],cx2[2][i]);
				printf("r0, r1, mu = %f %f %f\n",cr1[i],cr2[i],at[i]);
			}
			if(shapetype[i] == 'B'){
				printf("Object: %i - Box\n",i+1); 
				printf("x,  y,  z  = %f %f %f\n",x[0][i],x[1][i],x[2][i]);
				printf("lx, ly, lz = %f %f %f\n",d[0][i],d[1][i],d[2][i]);
				printf("psi, theta, phi, mu = %f %f %f %f\n",psi[i],theta[i],phi[i],at[i]);
			}
		}
        }
/*
	calculate equations of the objects
*/
	for(ii=0;ii<m;ii++){		/*loop through the objects*/
		at[ii] /= 10.0;  /*mm to cm*/
/*
	calculate rotation matrix
*/
        cpsi = cos(psi[ii]*convert);
        spsi = sin(psi[ii]*convert);
        cphi = cos(phi[ii]*convert);
        sphi = sin(phi[ii]*convert);
        ctheta = cos(theta[ii]*convert);
        stheta = sin(theta[ii]*convert);

        b[0][0] = cphi*cpsi - sphi*ctheta*spsi;
        b[0][1] = cphi*spsi + sphi*ctheta*cpsi;
        b[0][2] = sphi * stheta;
        b[1][0] = -sphi * cpsi - cphi * ctheta * spsi;
        b[1][1] = -sphi * spsi + cphi * ctheta * cpsi;
        b[1][2] = cphi * stheta;
        b[2][0] = stheta * spsi;
        b[2][1] = -stheta * cpsi;
        b[2][2] = ctheta;
/*
	handle special case of box
*/
	if(shapetype[ii] == 'B'){
		for(i=0;i<3;i++){
			nxy[i][ii] = b[i][2];
			nxz[i][ii] = b[i][1];
			nyz[i][ii] = b[i][0];
		}

		cxy[0][ii] = -.5 * nxy[0][ii] * d[2][ii] + x[0][ii] + xo;
		cxy[1][ii] = -.5 * nxy[1][ii] * d[2][ii] + x[1][ii] + yo;
		cxy[2][ii] = -.5 * nxy[2][ii] * d[2][ii] + x[2][ii] + zo;

		cxy[3][ii] = .5 * nxy[0][ii] * d[2][ii] + x[0][ii] + xo;
		cxy[4][ii] = .5 * nxy[1][ii] * d[2][ii] + x[1][ii] + yo;
		cxy[5][ii] = .5 * nxy[2][ii] * d[2][ii] + x[2][ii] + zo;

		cxz[0][ii] = -.5 * nxz[0][ii] * d[1][ii] + x[0][ii] + xo;
		cxz[1][ii] = -.5 * nxz[1][ii] * d[1][ii] + x[1][ii] + yo;
		cxz[2][ii] = -.5 * nxz[2][ii] * d[1][ii] + x[2][ii] + zo;

		cxz[3][ii] = .5 * nxz[0][ii] * d[1][ii] + x[0][ii] + xo;
		cxz[4][ii] = .5 * nxz[1][ii] * d[1][ii] + x[1][ii] + yo;
		cxz[5][ii] = .5 * nxz[2][ii] * d[1][ii] + x[2][ii] + zo;

		cyz[0][ii] = -.5 * nyz[0][ii] * d[0][ii] + x[0][ii] + xo;
		cyz[1][ii] = -.5 * nyz[1][ii] * d[0][ii] + x[1][ii] + yo;
		cyz[2][ii] = -.5 * nyz[2][ii] * d[0][ii] + x[2][ii] + zo;

		cyz[3][ii] = .5 * nyz[0][ii] * d[0][ii] + x[0][ii] + xo;
		cyz[4][ii] = .5 * nyz[1][ii] * d[0][ii] + x[1][ii] + yo;
		cyz[5][ii] = .5 * nyz[2][ii] * d[0][ii] + x[2][ii] + zo;

		dxy[0][ii] = nxy[0][ii]*cxy[0][ii]+nxy[1][ii]*cxy[1][ii]+nxy[2][ii]*cxy[2][ii];
		dxy[1][ii] = nxy[0][ii]*cxy[3][ii]+nxy[1][ii]*cxy[4][ii]+nxy[2][ii]*cxy[5][ii];

		dxz[0][ii] = nxz[0][ii]*cxz[0][ii]+nxz[1][ii]*cxz[1][ii]+nxz[2][ii]*cxz[2][ii];
		dxz[1][ii] = nxz[0][ii]*cxz[3][ii]+nxz[1][ii]*cxz[4][ii]+nxz[2][ii]*cxz[5][ii];

		dyz[0][ii] = nyz[0][ii]*cyz[0][ii]+nyz[1][ii]*cyz[1][ii]+nyz[2][ii]*cyz[2][ii];
		dyz[1][ii] = nyz[0][ii]*cyz[3][ii]+nyz[1][ii]*cyz[4][ii]+nyz[2][ii]*cyz[5][ii];

		continue;
	}
/*
	generate initial ellipsoid matrix
*/
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
                        a[i][j] = 0.0;
                }
                a[i][i] = 1.0 / (d[i][ii] * d[i][ii]);
/*start DSP*/
			/* unrotated cylinder has infinite radius in x dir */
 		if (shapetype[ii] == 'Y') a[0][i] = 0.0;

			/* axis of unrotated cone is along x axis */
		if (shapetype[ii] == 'C') a[0][0] *= -1.0;
/*end DSP*/
        }
/*
	form [b^t]ab
*/
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
                        c[i][j] = 0.0;
			for(l=0;l<3;l++){
                                c[i][j] += a[i][l] * b[l][j];
                        }
                }
        }

	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
                        a[i][j] = 0.0;
			for(l=0;l<3;l++){
                                a[i][j] += b[l][i] * c[l][j];
                        }
                }
        }
/*
	compute coefficients of ellipse centered at the origin
*/
        g1 = a[0][0];
        g2 = a[1][1];
        g3 = a[2][2];
        g4 = (a[0][1] + a[1][0])*0.5;
        g5 = 0.5 * (a[0][2] + a[2][0]);
        g6 = 0.5 * (a[1][2] + a[2][1]);
/*start DSP*/
	if ((shapetype[ii] == 'E') || (shapetype[ii] == 'Y'))
	  g7 = -1.0;
	if (shapetype[ii] == 'C')
	  g7 = 0.0;		
/*end DSP*/

/*
	translate origin to [x[0,ii],x[1,ii],x[2,ii]]
*/
        q[0][ii] = g1;
        q[1][ii] = g2;
        q[2][ii] = g3;
        q[3][ii] = g4;
        q[4][ii] = g5;
        q[5][ii] = g6;
        x1 = x[0][ii] + xo;
        x2 = x[1][ii] + yo;
        x3 = x[2][ii] + zo;
        q[6][ii]=g7+g1*x1*x1+g2*x2*x2+g3*x3*x3+2.0
                   *(g4*x1*x2+g5*x1*x3+g6*x2*x3);
        q[7][ii] = -(g1*x1 + g4*x2 + g5*x3);
        q[8][ii] = -(g2*x2 + g4*x1 + g6*x3);
        q[9][ii] = -(g3*x3 + g5*x1 + g6*x2);

        }                  /*end ellipsoid loop*/
/*
	program branching
*/
	if(pmode == IMAGE){
		gen_image();
		goto out;
	}
/*
	initialization for projection generation
*/
	tphase *= pi/180.;
	iblk = 1;
/*
	set up fan and parallel geometries
*/
	if(algo != CT_PAR){
		for(i=0;i<n;i++){
			if(algo == CT_SPACE){
				t = wdiso * (i-cen);
				ggamma = atan(t/rs);
			}else{
				ggamma = (i-cen)*dgamma;
			}

			cosd[i] = rd*cos(ggamma) - rs;
			sind[i] = rd*sin(ggamma);

			coss[i] = rs;
			sins[i] = 0.0;
			adet[i] = ggamma;
		}
	}else{				/* parallel projections */
		for(i=0;i<n;i++){
			cosd[i] = rd - rs;
			sind[i] = (i-cen) * wdiso;

			coss[i] = rs;
			sins[i] = -(i-cen) * wdiso;
			adet[i] = asin(sind[i]/rs);
		}
	}
/*
	z's of source
*/
        dz = hs / nzs;            
        z = (-hs + dz) * 0.5;
	for(i=0;i<nzs;i++){
                zdiss[i] = z;
                z += dz;
        }
/*
	azimuthal source parameters: weights and positions
*/
	if(target != 90.0){
		target *= convert;
		ls = hs * cos(target) / sin(target);
	}

	for(j=0;j<n;j++){
		if(target != 90.0){
			xb = ws * cos(adet[j]);
			xc = fabs(ls*sin(adet[j]));
		}else{
			xb = ws;
			xc = 0.0;
		}
		fwzm = xb + xc;
		fwfm = fabs(xb-xc);

		as = (fwzm - fwfm)/2;
		bs = fwfm;
		cs = as*2 + bs;

		if(debug){
			printf("detector: %d\n",j);
			printf("\ths,ws,ls: %f %f %f\n",hs,ws,ls);
			printf("\txb,xc: %f %f\n",xb,xc);
			printf("\tfwzm,fwfm: %f %f\n",fwzm,fwfm);
			printf("\tas,bs,cs: %f %f %f\n",as,bs,cs);
		}

		dt = cs / nas;

		for(i=0;i<nas;i++){
			ii = nas*j + i;
			t = -cs/2 + dt/2 + dt * i;
			if(t < -bs/2){
				wsa[ii] = (t + bs/2 + as) / as;
			}else if(t > bs/2){
				wsa[ii] = (-t + bs/2 + as) / as;
			}else{
				wsa[ii] = 1.0;
			}
			if(debug) printf("\t%d %f %f\n",i,t,wsa[ii]);

			csa[ii] = cos(atan(t/rs));
			ssa[ii] = sin(atan(t/rs));
		}

		for(t=i=0;i<nas;i++){
			ii = nas*j + i;
			t += wsa[ii];
		}
		t = nas / t;
		for(i=0;i<nas;i++){
			ii = nas*j + i;
			wsa[ii] *= t;
			if(debug) printf("\t%d %f\n",i,wsa[ii]);
		}
	}

/*	
	z's of detector
*/
	pp /= 100;
        dz = hd / nzd;    
        z = (-hd + dz) * 0.5;
	z += zdet;
	dzwp = 0.0;
	for(i=0;i<nzd;i++){
		if(fabs(z-zdet) > pp * hd/2){ /* CR */
			dzw[i] = (hd/2 - fabs(z)) / (hd/2 * ( 1.0 - pp));
		}else{
			dzw[i] = 1.0;
		}
		dzwp += dzw[i];
                zdisd[i] = z;
                z += dz;
        }
/*	
	detector azimuthal parameters
*/

        wdp = wd * da / 100.0;   
        rdp = rd - rs;
        dt = wdp / nad;
        t = (-wdp + dt) * 0.5;
	for(i=0;i<nad;i++){
                cda[i] = cos(atan(t/rdp));
                sda[i] = sin(atan(t/rdp));
                t = t + dt;
        }

        nprod = nzs * nas * dzwp * nad;
        if(eproj == 1){
                sclog = log(nprod);
        }else{
                sclin = 1.0 / nprod;
        }
/*
	misc. setup for view loop
*/
	i1 = 0; i2 = n;			/* by pass speed trick */
	helixzp = helixz = 0.0;		/* axial postions of source & detector */
/*
	loop through all the views
*/
	for(ii=0;ii<k*nrot;ii++){

	proj = aproj;
/*
	axial move for helical scanning
*/
	if(!var){
		helixz = slide * ii / (float)k;
		if(ii >= k) helixz += zjump;
		if(tstudy){
			helixz += sin(tfreq*2.0*pi*ii/k + tphase) 
				* tgain;
		}
	}else{
		if(ii < vbreak*k){
			helixz = slide/2 * ii / (vbreak*k);
		}else if(ii < k - vbreak*k - 1){
			helixz = slide/2;
		}else if(ii < k){
			helixz = slide/2*(2+(ii-k+1)/(vbreak*k));
		}else{
			helixz = slide;
		}
	}
	if(mbct){
	  if(clamp > 0){
		if(fabs((slide/2 - helixz)) > clamp*slide/2){
			if(slide/2 > helixz){
				helixzp = clamp*(slide/2)*
				(rd - rs)/rs + slide/2;
				helixzp += -(1-clamp)*slide/2 + helixz;
			}else{
				helixzp = -clamp*(slide/2)*
				(rd - rs)/rs + slide/2;
				helixzp += -(1+clamp)*slide/2 + helixz;
			}
		}else{
			helixzp = (slide/2 - helixz) *
			(rd - rs)/rs + slide/2;
		}
	  }else{
		helixzp = (slide/2 - helixz) * -clamp * 
		(rd - rs)/rs + slide/2;
	  }
	}else{
		helixzp = helixz;
	}
	if(debug)printf("%f %f\n",(double)helixz,(double)helixzp);
/*
	view angle
*/
        ths = pi / 2.0 + (alpha / 180.0 * pi) + ii * dths;
/*
	loop though all the projection samples
*/
	if(speed){
		for(i=0;i<n;i++){
			proj[i] = 0.0;              
		}
		proj += i1;
		istate = 0;
	}
	for(i=i1;i<i2;i++){
		psum = 0.0;              /*projection sum*/
/*
	put in das skew
*/
	thsp = ths + skew/100 * (i - cen) / n * dths;

	cths = cos(thsp);
	sths = sin(thsp);
/*
	loop through all the angle blocks in the source
*/
	for(isa=0;isa<nas;isa++){
		isap = isa + i * nas;
		cb = coss[i]*csa[isap] - sins[i]*ssa[isap];
		sb = coss[i]*ssa[isap] + sins[i]*csa[isap];
		xs = cths * cb - sths * sb;
		ys = cths * sb + sths * cb;

		/*
		xs = cths*csa[isa] - sths*ssa[isa];
		ys = sths*csa[isa] + cths*ssa[isa];
		*/

		xs2 = xs * xs;
		ys2 = ys * ys;
/*
	loop through all the z blocks in the source
*/
	for(isz=0;isz<nzs;isz++){
		hs = zdiss[isz] + helixz;
		hs2 = hs * hs;
/*
	loop through all the angle blocks in the detector
*/
	for(ida=0;ida<nad;ida++){
		cb = cosd[i]*cda[ida] - sind[i]*sda[ida];
		sb = cosd[i]*sda[ida] + sind[i]*cda[ida];
		xd = -cths * cb + sths * sb;
		yd = -cths * sb - sths * cb;
		ax1 = xd - xs;
		ax2 = yd - ys;
		ax12 = ax1 * ax1;
		ax22 = ax2 * ax2;
/*
	loop through all the z blocks in the detector
*/
	for(idz=0;idz<nzd;idz++){
		hd = zdisd[idz] + helixzp;
		spv = 0.0;
		ax3 = hd - hs;
		ax32 = ax3 * ax3;
/*
	loop through all the shapes
*/
	if(!additive) tsort_init();

	for(j=0;j<m;j++){


	if(shapetype[j] == 'B'){		/* intersect boxes */
		ra[0] = xd - xs;
		rb[0] = xs;
		ra[1] = yd - ys;
		rb[1] = ys;
		ra[2] = hd - hs;
		rb[2] = hs;

		delt = nxy[0][j]*ra[0] + nxy[1][j]*ra[1] + nxy[2][j]*ra[2];

		if(delt == 0.0){
			delt = sqrt_huge;
		}else{
			delt = 1.0 / delt;
		}

		dotb = nxy[0][j]*rb[0] + nxy[1][j]*rb[1] + nxy[2][j]*rb[2];
		x1 = (dxy[0][j] - dotb) * delt;
		x2 = (dxy[1][j] - dotb) * delt;

		if(delt > 0){
			tnz = x1;
			tfz = x2;
		}else{
			tnz = x2;
			tfz = x1;
		}

		delt = nxz[0][j]*ra[0] + nxz[1][j]*ra[1] + nxz[2][j]*ra[2];

		if(delt == 0.0){
			delt = sqrt_huge;
		}else{
			delt = 1.0 / delt;
		}

		dotb = nxz[0][j]*rb[0] + nxz[1][j]*rb[1] + nxz[2][j]*rb[2];
		x1 = (dxz[0][j] - dotb) * delt;
		x2 = (dxz[1][j] - dotb) * delt;

		if(delt > 0){
			tny = x1;
			tfy = x2;
		}else{
			tny = x2;
			tfy = x1;
		}

		delt = nyz[0][j]*ra[0] + nyz[1][j]*ra[1] + nyz[2][j]*ra[2];


		if(delt == 0.0){
			delt = sqrt_huge;
		}else{
			delt = 1.0 / delt;
		}

		dotb = nyz[0][j]*rb[0] + nyz[1][j]*rb[1] + nyz[2][j]*rb[2];
		x1 = (dyz[0][j] - dotb) * delt;
		x2 = (dyz[1][j] - dotb) * delt;

		if(delt > 0){
			tnx = x1;
			tfx = x2;
		}else{
			tnx = x2;
			tfx = x1;
		}

		t1 = tnx;
		t2 = tfx;
		if(tny > t1) t1 = tny;
		if(tnz > t1) t1 = tnz;
		if(tfy < t2) t2 = tfy;
		if(tfz < t2) t2 = tfz;
		if(t1 < t2){
			ar2 = 1;
		}else{
			ar2 = -1;
		}
	}else{			/* intersect ellipsoids, cones & cylinders */
		b1j = q[0][j];
		b2j = q[1][j];
		b3j = q[2][j];
		b4j = q[3][j];
		b5j = q[4][j];
		b6j = q[5][j];
		b7j = q[6][j];
		b8j = q[7][j];
		b9j = q[8][j];
		b10j = q[9][j];
		a1 = b1j*ax12+b2j*ax22+b3j*ax32+2.0*(b4j*ax1*ax2+b5j*ax1*ax3
		       +b6j*ax2*ax3);
		a2 = b1j*xs*ax1+b2j*ys*ax2+b3j*hs*ax3
		       +b4j*(xs*ax2+ys*ax1)+b5j*(xs*ax3+hs*ax1)
		       +b6j*(ys*ax3+hs*ax2)
		       +b8j*ax1+b9j*ax2+b10j*ax3;
		a3 = b1j*xs2+b2j*ys2+b3j*hs2
		       +2.0*(b4j*xs*ys+b5j*xs*hs+b6j*ys*hs)+b7j
		       +2.0*(b8j*xs+b9j*ys+b10j*hs);
		ar2 = a2*a2 - a1*a3;
	}
		if(ar2 >= 0.0){
		  if (shapetype[j] == 'E') {
		    t1 = (-a2-sqrt(ar2))/a1;
		    t2 = (-a2+sqrt(ar2))/a1;
		  }
		  if ((shapetype[j] == 'C')||(shapetype[j] == 'Y')) {
		    /* find and sort t of intersections with cyl/cone and with bounding planes */
		    tc1 = (-a2-sqrt(ar2))/a1;
		    tc2 = (-a2+sqrt(ar2))/a1;
		    tp1 = 0.0;  tp2 = 0.0;
		    normB = norm[0][j]*ax1+norm[1][j]*ax2+norm[2][j]*ax3;
		    if (normB != 0.0) {
		      tp1 = (norm[0][j]*(cx1[0][j]-xs)+norm[1][j]*(cx1[1][j]-ys)+norm[2][j]*(cx1[2][j]-hs))/
			normB;
		      tp2 = (norm[0][j]*(cx2[0][j]-xs)+norm[1][j]*(cx2[1][j]-ys)+norm[2][j]*(cx2[2][j]-hs))/
			normB;
		    }
		    if (tp1 > tp2) { t1 = tp1; tp1 = tp2; tp2 = t1; } /* sort tp1,tp2   (t1 is a temp var)*/
		    if (tc1 > tc2) { t1 = tc1; tc1 = tc2; tc2 = t1; } /* sort tc1,tc2 */
		    /* see if tp1 and tp2 hit the cyl/cone */
		    tp1in = 0; tp2in = 0;
/* ARE b1j, b2j and b3j necessarily pos? */
		    /*if ((b1j < 0)||(b2j < 0)||(b3j < 0)) {
		      printf("b1j=%f  b2j=%f  b3j=%f\b",b1j,b2j,b3j);
		    }*/
		    if (normB != 0.0) {
		      xi[0] = xs+ax1*tp1;  xi[1] = ys+ax2*tp1;  xi[2] = hs+ax3*tp1;
		      tp1in = (b1j*xi[0]*xi[0]+b2j*xi[1]*xi[1]+b3j*xi[2]*xi[2]+2*b4j*xi[0]*xi[1]+
			   2*b5j*xi[0]*xi[2]+2*b6j*xi[1]*xi[2]+b7j+2*b8j*xi[0]+2*b9j*xi[1]+2*b10j*xi[2]<=0.0);
		      xi[0] = xs+ax1*tp2;  xi[1] = ys+ax2*tp2;  xi[2] = hs+ax3*tp2;
		      tp2in = (b1j*xi[0]*xi[0]+b2j*xi[1]*xi[1]+b3j*xi[2]*xi[2]+2*b4j*xi[0]*xi[1]+
			   2*b5j*xi[0]*xi[2]+2*b6j*xi[1]*xi[2]+b7j+2*b8j*xi[0]+2*b9j*xi[1]+2*b10j*xi[2]<=0.0);
		    }
		    if (tp1in && tp2in) {     /* entry & exit on planes */
		      t1 = tp1;
		      t2 = tp2;
		    }
		    if (!tp1in && !tp2in)     /* entry & exit on cyl/cone */
		      if (normB == 0.0) {       /* line is parallel to planes */
			xi[0] = xs+ax1*tc1;  xi[1] = ys+ax2*tc1;  xi[2] = hs+ax3*tc1;
			plane1 = norm[0][j]*(xi[0]-cx1[0][j]) + norm[1][j]*(xi[1]-cx1[1][j]) + 
			         norm[2][j]*(xi[2]-cx1[2][j]);
			plane2 = norm[0][j]*(xi[0]-cx2[0][j]) + norm[1][j]*(xi[1]-cx2[1][j]) + 
			         norm[2][j]*(xi[2]-cx2[2][j]);
			if (plane1 * plane2 < 0.0) {
			  t1 = tc1;
			  t2 = tc2;
			} else {
			  t1 = t2 = 0.0;
			}
		      } else
			if ((tp1<tc1)&&(tc1<tp2) && (tp1<tc2)&&(tc2<tp2)) {
			  t1 = tc1;
			  t2 = tc2;
			} else {
			  t1 = t2 = 0.0;
			}
		    if (tp1in && !tp2in) {   /* entry on plane, exit on cyl/cone */
		      t1 = tp1;
		      if ((tc1 > tp1)&&(tc2 > tp1))
			t2 = MIN(tc1,tc2);
		      else
			t2 = MAX(tc1,tc2);
		    }
		    if (!tp1in && tp2in) {   /* entry on cyl/cone, exit on plane */
		      t2 = tp2;
		      if ((tc1 < tp2)&&(tc2 < tp2))
			t1 = MAX(tc1,tc2);
		      else
			t1 = MIN(tc1,tc2);
		    }
		  }
		  if (t2 < t1) {
		    /* printf("BUG!\n"); exit(1); */
		    t1 = t2;
		  }
		  if(additive){
			  spv += sqrt(ax12+ax22+ax32) * (t2-t1) * at[j];
		  }else{
			  tsort_insert(t1,t2,j);
		  }
		} 
        }                /*end loop through shapes */

	if(!additive) spv = tsort_attenuate() * sqrt(ax12+ax22+ax32);
/*end DSP*/

        if(eproj){
                psum += exp(-1.0 * spv) * dzw[idz] * wsa[isap];
        }else{
                psum += spv * dzw[idz] * wsa[isap];
        }

        }                /*end detector z block loop*/
	}                /*end detector angle block loop*/
	}                /*end source z block loop*/
	}                /*end source angle block loop*/

	if(noise){
		for(a1=l=0;l<7;l++){
			a1 += drand48();
		}
                psum += (a1 - 3.5)  /7.0;
	}

        if(eproj){
                *proj++ = psum = -log(psum) + sclog;
        }else{
                *proj++ =  psum = psum * sclin;
        }

/*
	set up speed trick for next view
*/
	if(speed && !istate && psum != 0){
		istate = 1;
		i1 = i;
	}

	if(speed && istate && psum == 0) break;
/*
	end projection sample loop
*/
	}                
/*
	output projection
*/
			/* trash the first nasym projections */
	for(i=0;i<ndeta;i++)temp[i] = aproj[nasym+i];
	parse_to_can_floatv(temp,ndeta);
	parse_wblock(ou,iblk,(char *)temp,nblk);
	iblk += nblk;

	if(pnview)if(ii%pnview==0)printf("views done : %i\n",ii);

/*
	update speed parameters for next iteration
*/
	if(speed){
		i1 *= .9;
	}		
/* 
	end projection loop 
*/
	}

/*
	clean up and exit
*/
out:	if(print)parse_pdt("---stop time");
        exit(0);
}

