/*
helical - perform helical interpolations

notes:

1. most of the algorithms are described in the paper: Crawford, C. R.,
and King, K. F., "Computed tomography scanning with simultaneous patient
translation," "Medical Physics," Vol. 17, No. 6, November/December 1990,
pp. 967-982.

2. the fast reconstruction of overlapping slices (fast=1) is described
in Crawford, C. R., King, K. F., and Hu, H., "Fast reconstruction of
overlapping slices using helical CT," presented at the "78th Scientific
Assembly of The Radiology Society of North America," Chicago, Illinois,
November 28 - December 3, 1993.  Abstract appeared in "Radiology," Vol.
189(P), November 1993, p. 217.

3. the HJ (jiang hsieh) algorithm is described in Hsieh, Jiang, "A
general approach to the reconstruction of x-ray helical computed
tomography," Med. Phys. 23(2), pp. 221-229, February 1996.

*/

static char rcsid[] = "$Id: helical.c,v 1.4 2008/03/07 18:55:50 carl Exp $";

#include	<math.h>
#include 	<parse.h>
#include 	"ct.h"

static char *ifn="hproj.da";	/* input file name */
static char *ofn="proj.da";	/* output file name */
static int print=1;		/* flag for printing program info */
static float oangle=45; 	/* overscan angle*/
static int debug=0;		/* 1 = debug mode */
static float uangle=45;		/* underscan angle */
static center=0;		/* 1=use centered data */
static float sextra=20;		/* extra data for segments */
static int cosine=0;		/* 1=mutliply by cosine of detector angle */
static int pcos=1;		/* power of cosine weighting */
static int fast=0;		/* 1=fast mode - filter w/ weight derivative */
static int k0=0;		/* first projection; center can overide */
static int feather=10;		/* number of feather channels */
static float aangle=3;		/* angle (degrees) asymmetric fan */
static int amode=0;		/* algorithm mode; allow permutations */
MODE imode=HE;			/* interplotaion modes */
static float phi=0.0;		/* offset angle for HA (degrees) */
static float cl_scale=1.0;	/* command line scale factor */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"phi",PARSE_FLOAT,(PARSE_ARG)&phi,"offset angle for HA"},
	{"aangle",PARSE_FLOAT,(PARSE_ARG)&aangle,"asymmetric angle"},
	{"aangle",PARSE_VALUE(HA),(PARSE_ARG)&imode},
	{"oangle",PARSE_FLOAT,(PARSE_ARG)&oangle,"overscan angle"},
	{"oangle",PARSE_VALUE(OS),(PARSE_ARG)&imode},
	{"uangle",PARSE_FLOAT,(PARSE_ARG)&uangle,"undercan angle"},
	{"uangle",PARSE_VALUE(US),(PARSE_ARG)&imode},
	{"extra",PARSE_FLOAT,(PARSE_ARG)&sextra},
	{"extra",PARSE_VALUE(HSE),(PARSE_ARG)&imode},
	{"k0",PARSE_INT,(PARSE_ARG)&k0,"first projection (-c override)"},
	{"pcos",PARSE_INT,(PARSE_ARG)&pcos,"power of cosine weighting"},
	{"pcos",PARSE_SET,(PARSE_ARG)&cosine,},
	{"feather",PARSE_INT,(PARSE_ARG)&feather,"# feather channels"},
	{"amode",PARSE_INT,(PARSE_ARG)&amode,"algorithm mode"},
	{"scale",PARSE_FLOAT,(PARSE_ARG)&cl_scale,"scale factor"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'i',(int *)&imode,PARSE_FLAG_VALUE(HI),"helical interpolative (HI)"},
	{'o',(int *)&imode,PARSE_FLAG_VALUE(OS),"overscan (OS)"},
	{'e',(int *)&imode,PARSE_FLAG_VALUE(HE),"helical extrapolative (HE)"},
	{'h',(int *)&imode,PARSE_FLAG_VALUE(HH),"helical halfscan (HH)"},
	{'s',(int *)&imode,PARSE_FLAG_VALUE(HS),"segment (halfscan) (HS)"},
	{'u',(int *)&imode,PARSE_FLAG_VALUE(US),"underscan (US)"},
	{'H',(int *)&imode,PARSE_FLAG_VALUE(HH2),"helical halfscan (2nd emb.) (HH2)"},
	{'t',(int *)&imode,PARSE_FLAG_VALUE(FS),"pass through mode (FS)"},
	{'S',(int *)&imode,PARSE_FLAG_VALUE(HSE),"segment (halfscan with extra) (HSE)"},
	{'a',(int *)&imode,PARSE_FLAG_VALUE(HA),"asymmetric detector (HA)"},
	{'j',(int *)&imode,PARSE_FLAG_VALUE(HJ),"jiang hsieh (HJ)"},
	{'d',&debug,PARSE_FLAG_SET,"debug mode"},
	{'f',&fast,PARSE_FLAG_SET,"fast mode - weight derivatives"},
	{'c',&center,PARSE_FLAG_SET,"use centered data"},
	{'C',&cosine,PARSE_FLAG_SET,"multiply by cosine of detector"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(int argc,char **argv)
{
	float rsc;		/*source to center distance*/
	int k;			/*number of projections - from header*/
	int n;			/*# of samples per projection - from header*/
	float r;		/*radius of scan circle*/
	int nrot;		/*# 360 degree rotations (from header)*/
	float alpha0;		/* projection starting angle */
	float dbeta;		/* bump between projections */
	float scale;		/* scale factor for back*/
	float wscale;		/* weight scale */
	int algo;		/* fan or parallel */
	int k_out;		/* # projections generated */
	int k_in;		/* number input projections */
	float *projp;		/* projection pointers */
	float *proj;		/* projection arrays */
	float *alpha;		/* array of detector angles */
	float beta_out;		/* offset angle to first projection */
	float convert;		/* pi / 180 */
	MODE imodep;		/* imode from input data */
	float slide;		/*z-slide [mm] per 360 rotation*/
	float cen;		/* center detector */
	float wdiso;		/* width of detector at iso-center */
	float alphaw;		/* anglular extent of feathering */
	int iblk,nblk,oblk;	/* input/output block pointers and counters */
	FILE *iu,*ou;		/* input/output file descriptors */
	float dgamma;		/* angle subtended by one detector */
	int nasym;		/* number of asymmetric channels */
	HU hu;			/* header for input and output files*/
	float delta;		/* half the fan angle */
	float beta;		/* angle of view */
	float betap;		/* angle of view offset by pi */
	float betah;		/* angle of view offset by pi and twice fan angle */
	int j,i,kp,ii;		/* loop counters and intermediate values */
	int rot_need;		/* number of rotations required */
	float w,w1,w2,w3;	/* projection weights */
	float w1p,w2p,w3p,w4p;	/* more projection weights */
	float f,f1;		/* feathering weights */
	float p0,aa,alphac;	/* intermediate angles */
	float alphad,alphae;	/* more intermediate angles */
	float pi;		/* ... in you face */
	float x;		/* distance on flat detector array */
	int nopar;		/* 1=parallel projections not supported */
/*
	parse command line
*/
	parse_all_basic(".helical","HELICAL",argc,argv);
/*
	open files
*/
	ifn = parse_suffix_malloc(ifn,"da");
	ofn = parse_suffix_malloc(ofn,"da");
	iu = parse_fopen(ifn,"r");
	ou = parse_fopen(ofn,"w");
/*
	get header information
*/
	ct_read_header(iu,&hu);

	n = hu.heads[CT_N];
	k_in = hu.heads[CT_K];
	algo = hu.heads[CT_ALGO];
	nrot = hu.heads[CT_ROT];		
	imodep = hu.heads[CT_MODE];		

	r = hu.headf[CT_R];
	rsc = hu.headf[CT_RSC];
	alpha0 = hu.headf[CT_ALPHA];
	dbeta = hu.headf[CT_DBETA];
	scale = hu.headf[CT_SCALE];
	slide = hu.headf[CT_SLIDE];
	cen = hu.headf[CT_CEN];
	wdiso = hu.headf[CT_WDISO];
/*
	set up buffers
*/
	nblk = BLOCK_FLOAT(n);
	i = nblk * 512 / sizeof(float);
	projp = parse_malloc_float(i);
	alpha = parse_malloc_float(i);
/*
	check parameter validity
*/
	if(imodep != FS)parse_er("helical: projections not in pass through mode");

	parse_pcheck(imodep,0,MAX_MODE,"helical: illegal projection mode");
        parse_pcheck(nrot,1,1024,"helical: bad rotations in header");
/*
	set up algorithm independent information
*/
        k = k_in / nrot;
        pi = M_PI;

	dgamma = atan(wdiso/rsc);

	if(algo == CT_PAR){
		for(i=0;i<n;i++){
			x = (i-cen) * wdiso;
			alpha[i] = asin(x/rsc);
		}
	}else if(algo == CT_SPACE){
		for(i=0;i<n;i++){
			x = (i-cen) * wdiso;
			alpha[i] = atan2(x,rsc);
		}
	}else if(algo == CT_ANGLE){
		for(i=0;i<n;i++){
			alpha[i] = (i-cen) * dgamma;
		}
	}else{
		parse_er("helical: internal error: can't set up for geometry");
	}

	delta = asin(r/rsc);
	alphaw = feather * dgamma;
	beta = dbeta * 0.5;		
	convert = pi / 180.0;
/*
	setup for the different modes
*/
        nopar = 0;			/* assume parallel supported */

	if(imode == HI){		/* 720 interpolative */
		k_out = 2*k;
		wscale = 1.0;
		rot_need = 2;
	}else if(imode == OS){	/* overscan */
		kp = oangle * convert / dbeta;
		k_out = k + kp;
		wscale = 1.0;
		rot_need = 2;
	}else if(imode == HE){	/* helical extrapolative 360 */
		k_out = k;
		wscale = 2.0;
		rot_need = 1;
	}else if(imode == HJ){	/* helical jiang hsieh 360 */
		k_out = k;
		wscale = 2.0;
		rot_need = 1;
		nopar = 1;
	}else if(imode == HH){	/* helical halfscan*/
		kp = k_out = k + 4* delta / dbeta + 0.5;
		wscale = 2.0;
		rot_need = 2;
		nopar = 1;
	}else if(imode == HS){	/* segment */
		k_out = kp = (pi + 2*delta) / dbeta;
		wscale = 2.0;
		rot_need = 1;
		nopar = 1;
	}else if(imode == US){	/* underscan */
		wscale = 1.0;
		k_out = k;
		rot_need = 1;
		nopar = 1;
		if(uangle*convert > pi/2 - delta)
			parse_err("underscan angle too large","");
	}else if(imode == HH2){	/* helical halfscan, embodiment 2*/
		kp = k_out = k + 4* delta / dbeta + 0.5;
		wscale = 2.0;
		rot_need = 2;
		nopar = 1;
	}else if(imode == FS){	/* pass through */
		k_out = k;
		wscale = 1.0;
		rot_need = 1;
	}else if(imode == HSE){	/* segment with extra data */
		k_out = kp = (pi + 2*delta + sextra*convert) / dbeta;
		wscale = 2.0;
		rot_need = 1;
		nopar = 1;
	}else if(imode == HA){	/* asymmetric array */
		parse_pcheck(amode,0,4,"helical: amode");
		aangle *= convert;
		nasym = aangle / dgamma;
		if(amode == 0 || amode == 3 || amode == 4){
			k_out = 2 * k;
			wscale = 2.0;
			rot_need = 2;
		}else{
			k_out = k;
			wscale = 1.0;
			rot_need = 1;
		}
	}

	wscale *= cl_scale;

	if(rot_need > nrot) parse_pri("*helical: rotations needed:",rot_need);
	if(nopar && algo == CT_PAR)
		parse_er("helical: can't use parallel projections");
/*
	set up for reading first projection
*/
	if(center) k0 = (nrot*k - k_out)/2;
	beta_out = k0 * dbeta;
	iblk = 1 + k0*nblk;
	if(k0 + k_out > k*nrot)parse_er("helical: insufficient projections");
/*
	print program status
*/
        if(print){
		parse_wtext("******************* helical *******************");
                parse_pdt("---start time");
		parse_rcsid_print(rcsid);
                parse_prs("input file:",ifn);
                parse_prs("output file:",ofn);
                parse_pri("n =",n);
                parse_pri("k (total) =",k_in);
                parse_pri("k (rot) =",k);
		parse_pri("k (out) =",k_out);
                parse_pri("nrot =",nrot);
		parse_prf("rsc =",rsc);
		parse_prf("r =",r);
		parse_prf("fan angle (degrees) =",delta/convert);
		parse_prf("center detector =",cen);
                parse_prf("wd (iso) =",wdiso);
		parse_prf("slide (mm) =",slide);
		parse_prf("scale =",cl_scale);
		parse_prt("center projection data",center);
		parse_pri("first projection (k0) =",k0);
		parse_prt("cosine modulaton",cosine);
		if(cosine)parse_pri("cosine power =",pcos);

		if(algo == CT_PAR){
			parse_wtext("parallel projections");
		}else if(algo == CT_SPACE){
			parse_wtext("fanbeam projections: equispace");
		}else{
			parse_wtext("fanbeam projections: equiangle");
		}

		ct_print_mode(imode);
		if(imode == HI){
			;
		}else if(imode == OS){
			parse_prf("oangle =",oangle);
		}else if(imode == HE){
			;
		}else if(imode == HJ){
			;
		}else if(imode == HH){
			;
		}else if(imode == HS){
			;
		}else if(imode == US){
			parse_prf("uangle =",uangle);
		}else if(imode == HH2){
			;
		}else if(imode == FS){
			;
		}else if(imode == HSE){
			parse_prf("extra =",sextra);
		}else if(imode == HA){
			parse_prf("aangle =",aangle/convert);
			parse_prf("asymmetric channels =",nasym);
			parse_pri("amode =",amode);
			parse_prf("phi (deg) =",phi);
		}
		if(imode == HE || imode == HA){
			parse_pri("feather channels =",feather);
		}
		parse_prt("fast recon mode",fast);
		parse_prt("debug mode",debug);
        }
/* 
	convert angles to radians
*/
	phi *= convert;
	sextra *= convert;
	uangle *= convert;
	oangle *= convert;
/*
	write header information
*/
	hu.heads[CT_N] = n;
	hu.heads[CT_K] = k_out;
	hu.heads[CT_ALGO] = algo;
	hu.heads[CT_ROT] = 1;		
	hu.heads[CT_MODE] = imode;		

	hu.headf[CT_R] = r;
	hu.headf[CT_RSC] = rsc;
	hu.headf[CT_ALPHA] = alpha0 + beta_out;
	hu.headf[CT_DBETA] = dbeta;
	hu.headf[CT_SCALE] = scale;
	hu.headf[CT_SLIDE] = slide;
	hu.headf[CT_CEN] = cen;
	hu.headf[CT_WDISO] = wdiso;

        ct_write_header(ou,&hu);

	oblk = 1;
/*
	loop through all the views
*/
	for(ii=0;ii<k_out;ii++){
/*
	read projection(s)
*/
	parse_rblock(iu,iblk,(char *)projp,nblk);
	parse_from_can_floatv(projp,n);
	iblk += nblk;
	proj = projp;		
/*
	handle different modes
*/
	if(imode == HI){		/* 720 interpolative */
		if(ii < k)
			scale = ((float) ii) / k;
		else
			scale =  2.0 - ((float) ii) / k;

		for(j=0;j<n;j++){
			if(debug)
				*proj++ = scale;
			else
				*proj++ *= scale;
		}
	}else if(imode == OS){	/* overscan */
		kp = oangle / dbeta;
		if(ii < kp || ii >= k){
			if(ii < kp){
				w = (ii + 0.5 ) / kp;
				w =  w * w * ( 3. - 2. * w);
			}else{
				w = (ii-k + 0.5 ) / kp;
				w =  w * w * ( 3. - 2. * w);
				w = 1.0 - w;
			}
			for(j=0;j<n;j++){
				if(debug)
					*proj++ = w;
				else
					*proj++ *= w;
			}
		}else if(debug == 1){
			for(j=0;j<n;j++){
				*proj++ = 1;
			}
		}
					/* jiang hsieh */
	}else if(imode == HJ){	
		for(i=0;i<n;i++){
			w1 = beta / ( pi - 2*alpha[i]);
			w2 = (2*pi - beta) / ( pi + 2*alpha[i]);

			if(beta <= pi - 2 * alpha[i]){
				w = w1;
			}else{
				w = w2;
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
					/* extrapolative 360 for fan */
	}else if(imode == HE && algo != CT_PAR){	
		alphac = (pi - beta) / 2;
		for(i=0;i<n;i++){
			if(fast == 0){
				w1 = (beta + 2*alpha[i]) / ( pi + 2*alpha[i]);
				w2 = (2*pi - beta - 2*alpha[i]) 
					/ ( pi - 2*alpha[i]);
			}else{
				w2 = 1.0 / (pi-2*alpha[i]);
				w1 = -1.0 / (pi+2*alpha[i]);
			}

			if(beta <= pi - 2 * delta){
				f = 1;
			}else if(beta >= pi + 2*delta){
				f = 0;
			}else if(alpha[i] < alphac - alphaw/2){
				f = 1;
			}else if(alpha[i] >= alphac + alphaw/2){
				f = 0;
			}else{
				f = (alpha[i] - alphac)/alphaw + .5;
				f = 1 - f;
				f = f * f *(3 - 2*f);
			}
			w = w1 * f + w2 * (1 - f);
			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
					/* extrapolative 360 for parallel */
	}else if(imode == HE && algo == CT_PAR){	
		alphac = pi - beta;
		alphad = beta;
		alphae = beta - 2*pi;

		for(i=0;i<n;i++){
			w1p = (alpha[i] + beta) / (2*alpha[i] - pi);
			w2p = (alpha[i] + beta) / (2*alpha[i] + pi);
			w3p = (2*pi - alpha[i] - beta) / (pi - 2*alpha[i]);
			w4p = (-2*pi + alpha[i] + beta) / (pi + 2*alpha[i]);

			if(beta <= delta){		/* w1 & w2 */
				if(alpha[i] < alphad - alphaw/2){
					w = w2p;
				}else if(alpha[i] >= alphad + alphaw/2){
					w = w1p;
				}else{
					f = (alpha[i] - alphad)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w2p * f + w1p * (1 - f);
				}
			}else if(beta <= pi - delta){	/* w2 */
				w = w2p;
			}else if(beta <= pi + delta){	/* w2 & w3 */
				if(alpha[i] < alphac - alphaw/2){
					w = w2p;
				}else if(alpha[i] >= alphac + alphaw/2){
					w = w3p;
				}else{
					f = (alpha[i] - alphac)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w2p * f + w3p * (1 - f);
				}
			}else if(beta <= 2*pi - delta){	/* w3 */
				w = w3p;
			}else{ 				/* w3 & w4 */
				if(alpha[i] < alphae - alphaw/2){
					w = w4p;
				}else if(alpha[i] >= alphae + alphaw/2){
					w = w3p;
				}else{
					f = (alpha[i] - alphae)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w4p * f + w3p * (1 - f);
				}
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}else if(imode == HS){	/* segment */
		for(i=0;i<n;i++){
			if(beta < 2*delta - 2 * alpha[i]){ 
				w = beta / ( 2*delta - 2*alpha[i]);
				w = w * w * (3.0 - 2 * w);
			}else if(beta < pi - 2*alpha[i]){ 
				w = 1.0;
			}else{
				w = (2*delta +pi-beta) / 
					(2*delta + 2*alpha[i]);
				w = w * w * (3.0 - 2 * w);
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}else if(imode == HSE){	/* segment with extra data */
		for(i=0;i<n;i++){
			if(beta < 2*delta - 2 * alpha[i] + sextra){ 
				w = beta/(2*delta-2*alpha[i]+sextra);
				w = w * w * (3.0 - 2 * w);
			}else if(beta < pi - 2*alpha[i]){ 
				w = 1.0;
			}else{
				w = (sextra + 2*delta +pi-beta) / 
					(sextra + 2*delta + 2*alpha[i]);
				w = w * w * (3.0 - 2 * w);
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}else if(imode == FS){	/* pass through */
		w = 1;
		for(i=0;i<n;i++){
			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}else if(imode == US){	/* underscan */
		if(beta < uangle){
			w = beta / uangle;
			w =  w * w * ( 3. - 2. * w);
			for(i=0;i<n;i++){
				if(debug)
					*proj++ = w;
				else
					*proj++ *= w;
			}
		}else if(beta > 2*pi-uangle){
			w = (2 * pi - beta) / uangle;
			w =  w * w * ( 3. - 2. * w);
			for(i=0;i<n;i++){
				if(debug)
					*proj++ = w;
				else
					*proj++ *= w;
			}
		}else if(beta >  pi - 2*delta - uangle &&
		    beta <  pi + 2*delta + uangle){
			for(i=0;i<n;i++){
				aa = alpha[i];
				if(beta >  pi - 2*aa - uangle &&
				  beta <  pi - 2*aa + uangle){
					w = fabs(beta-pi+ 2*aa) / uangle;
					w = 2. - w * w * ( 3. - 2. * w);
				}else{
					w = 1.0;
				}
				if(debug)
					*proj++ = w;
				else
					*proj++ *= w;
				
			}
		}else{
			w = 1.0;
			for(i=0;i<n;i++){
				if(debug)
					*proj++ = w;
				else
					*proj++ *= w;
			}
		}
	}else if(imode == HH){	/* helical halfscan*/
		p0 = pi + 2 * delta;
		for(i=0;i<n;i++){
			if(beta < 2*delta - 2 * alpha[i]){ /* R1 */
				w = beta / ( 2*delta - 2*alpha[i]);
				w = w * w * (3.0 - 2.0 * w);
				w1 = 1.0 + (beta - p0) / (2*pi);
				w *= w1;
			}else if(beta < 4*delta){ /* R2 */
				w =(4*delta-beta)/(2*(delta +alpha[i]));
				w = w * w * (3.0 - 2 * w);
				w1 = 1.0 + (beta - p0) / (2*pi);
				w2 =  1.+ (beta - p0)/(pi + 2*alpha[i]);
				w = w*w1 + (1-w)*w2;
			}else if(beta < pi - 2*alpha[i]){ /* R3 */
				w =  1. + (beta - p0)/(pi + 2*alpha[i]);
			}else if(beta < pi + 2*delta){ /* R4 */
				w = (p0-beta) /(2*delta + 2*alpha[i]);
				w = w * w * (3.0 - 2 * w);
				w1 =  1.+ (beta - p0)/(pi + 2*alpha[i]);
				w *= w1;
			}else if(beta < pi+4*delta-2*alpha[i]){ /* R5 */
				w = (beta - p0) / (2*(-alpha[i]+delta));
				w = w * w * (3.0 - 2 * w);
				w1 =  1.+ (p0 - beta)/(pi - 2*alpha[i]);
				w *= w1;
			}else if(beta < 2*pi){ /* R6 */
				w =  1. + (p0 - beta)/(pi - 2*alpha[i]);
			}else if(beta<2*pi+2*delta-2*alpha[i]){ /* R7 */
				w = (beta -2*pi) / (2*(delta-alpha[i]));
				w = w * w * (3.0 - 2 * w);
				w1 = 1.0 + (p0 - beta) / (2 * pi);
				w2 =  1.+ (p0 - beta)/(pi - 2*alpha[i]);
				w = w1*w + (1-w)*w2;
			}else if(beta < 2*pi + 4*delta){	/* R8 */
				w = (2*p0 -beta) / (2*(delta+alpha[i]));
				w = w * w * (3.0 - 2 * w);
				w1 = 1.0 + (p0 - beta) / (2 * pi);
				w *= w1;
			}else{
				w = 0;
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}else if(imode == HH2){	/* helical halfscan, embodiment 2*/
		p0 = pi + 2 * delta;
		for(i=0;i<n;i++){
			if(fast == 0){
				if(beta < 2*delta - 2 * alpha[i]){ /* R1 */
					w = 0;
				}else if(beta < 4*delta){ /* R2 */
					w1 = 1.0 + (beta - p0) / (2*pi);
					w2 =  1.+ (beta - p0)/(pi + 2*alpha[i]);
					w = w2;
				}else if(beta < pi - 2*alpha[i]){ /* R3 */
					w =  1. + (beta - p0)/(pi + 2*alpha[i]);
				}else if(beta < pi + 2*delta){ /* R4 */
					w1 =  1.+ (beta - p0)/(pi + 2*alpha[i]);
					w = w1;
				}else if(beta < pi+4*delta-2*alpha[i]){ /* R5 */
					w1 =  1.+ (p0 - beta)/(pi - 2*alpha[i]);
					w = w1;
				}else if(beta < 2*pi){ /* R6 */
					w =  1. + (p0 - beta)/(pi - 2*alpha[i]);
				}else if(beta<2*pi+2*delta-2*alpha[i]){ /* R7 */
					w1 = 1.0 + (p0 - beta) / (2 * pi);
					w2 =  1.+ (p0 - beta)/(pi - 2*alpha[i]);
					w = w2;
				}else if(beta < 2*pi + 4*delta){	/* R8 */
					w = 0;
				}else{
					w = 0;
				}
			}else{
				if(beta < 2*delta - 2 * alpha[i]){ /* R1 */
					w = 0;
				}else if(beta < 4*delta){ /* R2 */
					w =  -1.0/(pi + 2*alpha[i]);
				}else if(beta < pi - 2*alpha[i]){ /* R3 */
					w =  -1.0/(pi + 2*alpha[i]);
				}else if(beta < pi + 2*delta){ /* R4 */
					w =  -1.0/(pi + 2*alpha[i]);
				}else if(beta < pi+4*delta-2*alpha[i]){ /* R5 */
					w =  1.0/(pi - 2*alpha[i]);
				}else if(beta < 2*pi){ /* R6 */
					w =  1.0/(pi - 2*alpha[i]);
				}else if(beta<2*pi+2*delta-2*alpha[i]){ /* R7 */
					w =  1.0/(pi - 2*alpha[i]);
				}else if(beta < 2*pi + 4*delta){	/* R8 */
					w = 0;
				}else{
					w = 0;
				}
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
				/* asymmetric detector; stomp on projections */
	}else if(imode == HA && amode == 2){	
		for(i=0;i<n;i++){
			if(alpha[i] < -delta + aangle){
				w = 0;
			}else{
				w = 1;
			}
			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
				/* asymmetric detector; double projections */
	}else if(imode == HA && amode == 1){	
		for(i=0;i<n;i++){
			if(alpha[i] < -delta + aangle){
				w = 0;
			}else if(alpha[i] < -delta + aangle + alphaw){
				w = -(alpha[i] + delta - aangle - alphaw)/alphaw;
				w = 1 - w;
				w = w * w * ( 3 - 2 * w);
			}else if(alpha[i] < delta - aangle - alphaw){
				w = 1;
			}else if(alpha[i] < delta - aangle){
				w = -(alpha[i] - delta + aangle)/alphaw;
				w = 1 - w;
				w = 1 + w * w * ( 3 - 2 * w);
			}else{
				w = 2;
			}

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
							/* peep - editor tag */
				/* asymmetric detector - HA; fan */
	}else if(imode == HA && (amode == 0 || amode == 3 || amode == 4) 
	  && algo != CT_PAR){	
		betap = beta - pi + phi;
		betah = betap + 2 * delta;
		alphac = (pi - betap) / 2;
		for(i=0;i<n;i++){

					/* HI weights */
			w1 = (betap + pi) / (2 * pi);
			w2 = (3*pi - betap) / (2 * pi);
			if(1 || phi == 0.0){
				if(betap < pi + phi){
					w3 = w1;
				}else{
					w3 = w2;
				}
			}else{
				if(betap < pi + phi - alphaw/2){
					w3 = w1;
				}else if(betap > pi + phi + alphaw/2){
					w3 = w2;
				}else{
					f = (betap - pi - phi)/alphaw + .5;
					f = f * f *(3 - 2*f);
					w3= w2 * f + w1* (1-f);
				}
			}

			if(alpha[i] < -delta + aangle+alphaw){
				w3 = 0;
			}else{
				;
			}

			if(amode == 0){
						/* HE weights */
				w1 = (betap + 2*alpha[i]) / (pi + 2*alpha[i]);
				w2 = (2*pi - betap - 2*alpha[i]) / (pi - 2*alpha[i]);

				if(betap <= -alphaw/2){
					w = 0;
				}else if(betap <= alphaw/2){
					f = betap/alphaw + .5;
					f = f * f *(3 - 2*f);
					w = w1 * f;
				}else if(betap <= pi - 2 * delta){
					w = w1;
				}else if(betap <= pi + 2 * delta){
					if(alpha[i] < alphac - alphaw/2){
						w = w1;
					}else if(alpha[i] <= alphac + alphaw/2){
						f = (alpha[i] - alphac)/alphaw + .5;
						f = 1 - f;
						f = f * f *(3 - 2*f);
						w = w1 * f + w2 * (1 - f);
					}else{
						w = w2;
					}
				}else if(betap <= 2*pi - alphaw/2){
					w = w2;
				}else if(betap <= 2*pi + alphaw/2){
					f = (betap - 2*pi)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w2 * f;
				}else{
					w = 0;
				}
						/* HJ weights */
			}else if(amode == 3){
				w1 = betap / ( pi - 2*alpha[i]);
				w2 = (2*pi - betap) / ( pi + 2*alpha[i]);

				if(betap <= 0){
					w = 0;
				}else if(betap <= pi - 2 * alpha[i]){
					w = w1;
				}else if(betap <= 2*pi){
					w = w2;
				}else{
					w = 0;
				}
						/* HH2 weights */
			}else if(amode == 4){
				p0 = pi + 2 * delta;
				if(betah < 2*delta - 2 * alpha[i]){ /* R1 */
					w = 0;
				}else if(betah < 4*delta){ /* R2 */
					w1 = 1.0 + (betah - p0) / (2*pi);
					w2 =  1.+ (betah - p0)/(pi + 2*alpha[i]);
					w = w2;
				}else if(betah < pi - 2*alpha[i]){ /* R3 */
					w =  1. + (betah - p0)/(pi + 2*alpha[i]);
				}else if(betah < pi + 2*delta){ /* R4 */
					w1 =  1.+ (betah - p0)/(pi + 2*alpha[i]);
					w = w1;
				}else if(betah < pi+4*delta-2*alpha[i]){ /* R5 */
					w1 =  1.+ (p0 - betah)/(pi - 2*alpha[i]);
					w = w1;
				}else if(betah < 2*pi){ /* R6 */
					w =  1. + (p0 - betah)/(pi - 2*alpha[i]);
				}else if(betah<2*pi+2*delta-2*alpha[i]){ /* R7 */
					w1 = 1.0 + (p0 - betah) / (2 * pi);
					w2 =  1.+ (p0 - betah)/(pi - 2*alpha[i]);
					w = w2;
				}else if(betah < 2*pi + 4*delta){	/* R8 */
					w = 0;
				}else{
					w = 0;
				}
			}

					/* feather HI weights */
			if(alpha[i] < -delta + aangle){
				f1 = 1;
			}else if(alpha[i] < -delta + aangle + alphaw){
				f1 = -(alpha[i] + delta - aangle - alphaw)/alphaw;
				f1 = f1 * f1 * ( 3 - 2 * f1);
			}else if(alpha[i] < delta - aangle - alphaw){
				f1 = 0;
			}else if(alpha[i] < delta - aangle){
				f1 = (alpha[i] - delta + aangle)/alphaw;
				f1 = 1 + f1;
				f1 =  f1 * f1 * ( 3 - 2 * f1);
			}else{
				f1 = 1;
			}

			w = f1*w3 + (1-f1)*w;

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
				/* asymmetric detector - HA; parallel */
	}else if(imode == HA && amode == 0 && algo == CT_PAR){	
		betap = beta - pi;
		alphac = pi - betap;
		alphad = betap;
		alphae = betap - 2*pi;
		for(i=0;i<n;i++){

					/* HI weights */
			if(betap-alpha[i] < -pi ){ 
				w3 = (-pi - betap + alpha[i]) / (2 * pi);
			}else if(betap-alpha[i] < pi){
				w3 = (betap-alpha[i] + pi) / (2 * pi);
			}else{
				w3 = (3*pi - betap + alpha[i]) / (2 * pi);
			}

			if(alpha[i] < -delta + aangle+alphaw){
				w3 = 0;
			}else{
				;
			}

					/* HE weights */
			w1p = 0;
			w2p = (alpha[i] + betap) / (2*alpha[i] + pi);
			w3p = (2*pi - alpha[i] - betap) / (pi - 2*alpha[i]);
			w4p = 0;

			if(betap <= -delta){		/* w1 */
				w = 0;
			}else if(betap <= delta){	/* w1 & w2 */
				if(alpha[i] < alphad - alphaw/2){
					w = w2p;
				}else if(alpha[i] >= alphad + alphaw/2){
					w = w1p;
				}else{
					f = (alpha[i] - alphad)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w2p * f + w1p * (1 - f);
				}
			}else if(betap <= pi - delta){	/* w2 */
				w = w2p;
			}else if(betap <= pi + delta){	/* w2 & w3 */
				if(alpha[i] < alphac - alphaw/2){
					w = w2p;
				}else if(alpha[i] >= alphac + alphaw/2){
					w = w3p;
				}else{
					f = (alpha[i] - alphac)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w2p * f + w3p * (1 - f);
				}
			}else if(betap <= 2*pi - delta){	/* w3 */
				w = w3p;
			}else if(betap <= 2*pi + delta){	/* w3 & w4 */
				if(alpha[i] < alphae - alphaw/2){
					w = w4p;
				}else if(alpha[i] >= alphae + alphaw/2){
					w = w3p;
				}else{
					f = (alpha[i] - alphae)/alphaw + .5;
					f = 1 - f;
					f = f * f *(3 - 2*f);
					w = w4p * f + w3p * (1 - f);
				}
			}else{
				w = 0;
			}

					/* feather HI weights */
			if(alpha[i] < -delta + aangle){
				f1 = 1;
			}else if(alpha[i] < -delta + aangle + alphaw){
				f1 = -(alpha[i] + delta - aangle - alphaw)/alphaw;
				f1 = f1 * f1 * ( 3 - 2 * f1);
			}else if(alpha[i] < delta - aangle - alphaw){
				f1 = 0;
			}else if(alpha[i] < delta - aangle){
				f1 = (alpha[i] - delta + aangle)/alphaw;
				f1 = 1 + f1;
				f1 =  f1 * f1 * ( 3 - 2 * f1);
			}else{
				f1 = 1;
			}

			w = f1*w3 + (1-f1)*w;

			if(debug)
				*proj++ = w;
			else
				*proj++ *= w;
		}
	}
/*
	perform other projection weighting
*/
	for(proj=projp,i=0;i<n;i++){
		*proj++ *= wscale;
	}

	if(cosine){
		for(proj=projp,i=0;i<n;i++){
			proj[i] *= pow(cos(alpha[i]),(double)pcos);
		}
	}
/*
	write out projection
*/
	parse_to_can_floatv(projp,n);
	parse_wblock(ou,oblk,(char *)projp,nblk);
	oblk += nblk;
/*
	update for next projection and end view loop
*/
	beta += dbeta;
}
/*
	clean up and exit
*/
        if(print==1)parse_pdt("---stop time");
        exit(0);
}
