/*
mtf - calculate MTF from mi-file
*/

static char rcsid[] = "$Id: mtf.c,v 1.7 2008/03/07 18:55:50 carl Exp $";

#include <parse.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "ct.h"

#define P 128			/* points in final mtf */

static int N;			/* size of image subset */
static int M;			/* log_2 of N */
static int npic;		/* size of image subset from command line */
static char *ifn="pic.mi";	/* input file name */
static char *ofn;		/* output file name */
static int print=1;		/* flag for printing program info */
static int debug=0;		/* print debug information */
static float fov=100;		/* field-of-view (mm) */
static int xo,yo;		/* x- and y-center, default auto calc */
static int xos=0,yos=0;		/* xo= and yo= set */
static char *ffn="mtff.da";	/* frequency file name */
static char *sfn="subset.mi";	/* image subset filename */
static char *fftfn="fft.da";	/* mag(fft(subset)) */
static char *afn="ave.da";	/* ave from circular integration */
static char *dcfn="dc.da";	/* w/o DC */
static float aang=180;		/* average range angle (degrees) */
static float cang=0;		/* center of average range angle (degrees) */
static int baseline=1024;	/* ... of image (HU) */
static float f_max=10;		/* max frequency (lp/cm) */
static int dcremove=0;		/* 1=remove DC */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file (default from ifn)"},
	{"fov",PARSE_FLOAT,(PARSE_ARG)&fov,"field of view (mm)"},
	{"xo",PARSE_INT,(PARSE_ARG)&xo,"x-origin, default=nx/2"},
	{"yo",PARSE_INT,(PARSE_ARG)&yo,"y-origin, default=ny/2"},
	{"aang",PARSE_FLOAT,(PARSE_ARG)&aang,"average angle (degrees)"},
	{"cang",PARSE_FLOAT,(PARSE_ARG)&cang,"center angle (degrees)"},
	{"fmax",PARSE_FLOAT,(PARSE_ARG)&f_max,"max frequency (lp/cm)"},
	{"baseline",PARSE_INT,(PARSE_ARG)&baseline,"image baseline"},
	{"npic",PARSE_INT,(PARSE_ARG)&npic,"size of sub-image; default=input"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'d',&debug,PARSE_FLAG_INC,"enable debugging output"},
	{'D',&dcremove,PARSE_FLAG_SET,"remove DC"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(int argc,char **argv)
{
	short int *ip;		/* input image */
	int nx,ny;		/* size of input image */
	short int *sub;		/* image subset */
	float *a;		/* real part of 2d fft */
	float *b;		/* imaginary part of 2d fft */
	int i,j,k,l;		/* loopers and int pointers */
	float pi;		/* don't eat, lose weight */
	float convert;		/* degrees to radians */
	float ix,iy;		/* frequencies */
	float w,w1;		/* weights */
	float y2,x2;		/* powers of frequencies */
	float omega;		/* magnitude of frequency */
	float *wt;		/* sum of weights for circle */
	float *sum;		/* sum of circle */
	int iomega;		/* floor(omega) */
	int iomega1;		/* floor(omega) + 1 */
	int circle;		/* number of circles */
	float theta;		/* polar angle */
	float gain;		/* used for mtf normalization */
	float df,dfo;		/* frequency steps */
	float *mtf;		/* final mtf */
	float f;		/* freq */
	int fi;			/* integer freq */
	float f50,f10;		/* freq at 50% and 10% */

			/* stuff for removal of DC */
	int nsub;		/* points to replace */
	int nuse;		/* adjacent points to use in fit */
	float sx2,sx4,s0,s2;	/* sums */
	float a0,a2;		/* averages */
	float xx;		/* locations */
/*
	parse command line and massage parameters
*/
	parse_all_basic(".mtf","MTF",argc,argv);

	xos = parse_opexist("xo");
	yos = parse_opexist("yo");

	sfn = parse_suffix_malloc(sfn,"mi");

	if(!parse_opexist("of")){
		ofn = parse_malloc_char(strlen(ifn)+1);
		strcpy(ofn,ifn);
		parse_remove_suffix(ofn);
	}
	ofn = parse_suffix_malloc(ofn,"da");

	pi = M_PI;
	convert = pi / 180.0;
/*
	open picture and get center NxN
*/
	ip = parse_rmi(ifn,NULL,&nx,&ny);

	if(parse_opexist("npic")){
		N = npic;
	}else{
		if(nx != ny){
			fprintf(stderr,"mtf: can't autoselect subset. image not square\n");
			fprintf(stderr,"nx: %d, ny %d\n",nx,ny);
			exit(1);
		}
		N = nx;
	}

	{			/* compute fft size and make sure power of 2 */

	float am;
	int nfft;

        am = log((float)(N))/log(2.0);
	M = am;
	if(am-M > 0.0)M = M + 1;
        nfft = pow(2.0,(float) M);
	if(nfft != N){
		parse_pri("**mtf: image subset not a power of two:",N);
	}

	}

	circle = N/2 * sqrt(2.) * 1.1;
	wt = parse_malloc_float(circle);
	sum = parse_malloc_float(circle);
	mtf = parse_malloc_float(P);
	a = parse_malloc_float(N*N);
	b = parse_malloc_float(N*N);
	sub = parse_malloc_short(N*N);

	if(xos == 0) xo = nx / 2;
	if(yos == 0) yo = ny / 2;
	xo -= N/2;
	yo -= N/2;

	if(xo < 0 || yo < 0 || xo+N > nx || yo+N > ny){
		parse_er("mtf: subset not contained in input image");
	}

	k = yo * nx + xo;
	for(l=i=0;i<N;i++){
		for(j=0;j<N;j++){
			sub[l] = ip[k++];
			a[l] = sub[l] - baseline;
			b[l] = 0.0;
			l++;
		}
		k += nx - N;
	}

	xo += N/2;		/* restore for printing */
	yo += N/2;

	if(debug) parse_wmi(sfn,sub,N,N);
/*
	print program status
*/
        if(print){
		parse_wtext("******************* mtf *******************");
                parse_pdt("---start time");
		parse_rcsid_print(rcsid);
                parse_prs("input file:",ifn);
                parse_prs("output file:",ofn);
                parse_prs("frequency file:",ffn);
                parse_pri("nx =",nx);
                parse_pri("ny =",ny);
                parse_pri("xo =",xo);
                parse_pri("yo =",yo);
                parse_pri("npic =",N);
                parse_prf("fov (mm) =",fov);
                parse_prf("aang (deg) =",aang);
                parse_prf("cang (deg) =",cang);
                parse_prf("f_max (lp/cm) =",f_max);
                parse_pri("baseline =",baseline);
                parse_prt("DC removal",dcremove);
		if(debug){
			parse_prs("subset:",sfn);
			parse_prs("fft:",fftfn);
			parse_prs("ave:",afn);
			parse_prs("w/o dc:",dcfn);
		}
        }
/*
	force 2d fft into center
*/
	for(l=i=0;i<N;i++){
		for(j=0;j<N;j++){
			if((i+j)%2){
				a[l] = -a[l];
				b[l] = -b[l];
			}
			l++;
		}
	}
/*
	take 2d fft and get magnitude
*/
	parse_fft2d(a,b,M,M,0);

	for(l=i=0;i<N;i++){
		for(j=0;j<N;j++){
			a[l] = hypot(a[l],b[l]);
			l++;
		}
	}

	if(debug) parse_wda(a,N,N,fftfn);
/*
	average along concentric circles 
*/
	cang *= convert;
	aang *= convert;

	parse_zrbuf(wt,circle);
	parse_zrbuf(sum,circle);

	for(k=0;k<N;k++){
		iy = k - N/2;
		y2 = iy * iy;

		for(l=0;l<N;l++){
			ix = l - N/2;
			if(ix == 0){
				if(iy == 0){
					theta = 0;
				}else{
					theta = pi/2 - cang;
				}
			}else{
				theta = atan2(iy,ix) - cang;
			}
			if(fabs(theta) > aang && fabs(theta+pi) > aang) continue;

			x2 = ix * ix;
			omega = sqrt(x2 + y2);
			iomega = floor(omega);
			iomega1 = iomega + 1;
			w = omega - iomega;
			w1 = 1 - w;
			i = k*N + l;

			wt[iomega] += w1;
			sum[iomega] += w1 * a[i];

			wt[iomega1] += w;
			sum[iomega1] += w * a[i];
		}
	}

	for(i=0;i<circle;i++){
		if(wt[i] == 0.0){
			sum[i] = 0.0;
		}else{
			sum[i] /= wt[i];
		}
	}

	if(debug) parse_wda(sum,circle,1,afn);
/*
	get rid of dc
*/
	if(dcremove){
		nsub = 3;
		nuse = 5;
		sx2 = 0.;
		sx4 = 0.;
		s0 = 0.;
		s2 = 0.;

		for(i=0;i<nuse;i++){
			xx = i + nsub;
			x2 = xx*xx;
			sx2 = sx2+x2;
			sx4 = sx4+x2*x2;
			s0 = s0+sum[i+nsub];
			s2 = s2+sum[i+nsub]*x2;
		}

		a0 = (s0*sx4-sx2*s2)/(nuse*sx4-sx2*sx2);
		a2 = (nuse*s2-sx2*s0)/(nuse*sx4-sx2*sx2);

		for(i=0;i<nsub;i++){
			xx = i;
			sum[i] = a0+a2*xx*xx;
		}

		if(debug) parse_wda(sum,circle,1,dcfn);
	}
/*
	normalize interpolate and save
*/
	gain = sum[0];

        if(gain == 0) parse_er("mtf: mtf maximum is zero");

        for(i=0;i<circle;i++){
        	sum[i] /= gain;
        }
/*
	interpolate and save
*/
	dfo = 1.0 / (fov / 10 / nx * N);
        df = f_max / (P - 1);

        f50 = f10 = -1;

        for(i=0;i<P;i++){
		f = i * df / dfo;
		fi = floor(f);
		w = f - fi;
		w1 = 1 - w;
        	
        	if(fi >= circle - 2){
        		mtf[i] = 0.0;
        	}else{
        		mtf[i] = w1*sum[fi] + w*sum[fi+1];
		}
		if(mtf[i] >= .5) f50 = f;
		if(mtf[i] >= .1) f10 = f;
        }

        if(print){
		if(f50 == -1){
			parse_wtext("50% (lp/cm) = NOT REACHED");
		}else{
			parse_prf("50% (lp/cm) =",f50*dfo);
		}
		if(f10 == -1){
			parse_wtext("10% (lp/cm) = NOT REACHED");
		}else{
			parse_prf("10% (lp/cm) =",f10*dfo);
		}
	}

	parse_wda(mtf,P,1,ofn);
/*
	fill and save frequency vector
*/
        for(i=0;i<P;i++){
        	mtf[i] = i * df;
        }
	parse_wda(mtf,P,1,ffn);
/*
	clean up and exit
*/
        if(print==1)parse_pdt("---stop time");
        return(0);
}
