/*
rebin - rebin fan-beam projections into parallel beam
*/

static char rcsid[] = "$Id: rebin.c,v 1.4 2008/03/07 18:55:50 carl Exp $";

#include	<math.h>
#include 	<parse.h>
#include 	"ct.h"

static char *ifn="proj.da";	/* input file name */
static char *ofn="pproj.da";	/* output file name */
static int print=1;		/* flag for printing program info */
static int k;			/*number of projections - from header*/
static int debug=0;		/* print debug information */
static int aio=8;		/* azimuthal interpolation order */
static int rio=8;		/* radial interpolation order */
static int intqr=256;		/* interpolation quantization steps radial */
static int intqa=256;		/* interpolation quantization steps azimuthal */
static int iso=0;		/* 0=normal 1=no 2=only iso interpolation */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"aio",PARSE_INT,(PARSE_ARG)&aio,"azimuthal interpolation order"},
	{"rio",PARSE_INT,(PARSE_ARG)&rio,"radial interpolation order"},
	{"aq",PARSE_INT,(PARSE_ARG)&intqa,"azimuthal quantization"},
	{"rq",PARSE_INT,(PARSE_ARG)&intqr,"radial quantization"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'d',&debug,PARSE_FLAG_INC,"enable debugging output"},
	{'i',&iso,PARSE_FLAG_VALUE(1),"no isocenter interpolation"},
	{'I',&iso,PARSE_FLAG_VALUE(2),"only isocenter interpolation"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

/*
	build_itbl - build interpolation kernel
*/
static void build_itbl(float *kern,int n,int offset,int intq)
{
	int i,j,m;
	float *kp=kern;
	float *kpp;
	float sum;
	float c;
	float beta;

	for(m=0;m<intq;m++){
		beta = 1.0 / intq * m + offset;
		kpp = kp;
		sum = 0;
		for(i=0;i<n;i++){
			c = 1;
			for(j=0;j<n;j++){
				if(i == j) continue;
				c *= (beta - j) / (i - j);
			}
			if(debug) printf("%4.3f ",c);
			sum += c;
			*kp++ = c;
		}
		if(debug){
			printf("%10.9f ",sum);
			putchar('\n');
		}
	}
}

/*
	smod - super modulo k 
*/
static int smod(int i)
{
	while(i<0) i += k;
	while(i>=k) i -= k;
	return(i);
}

int
main(int argc,char **argv)
{
	float rsc;		/*source to center distance*/
	int nf;			/*# of samples per fan projection - from header*/
	int np;			/*# of samples per parallel projection */
	float r;		/*radius of scan circle*/
	int nrot;		/*# 360 degree rotations (from header)*/
	float beta0;		/* projection starting angle */
	float dbeta;		/* bump between projections */
	float scale;		/* scale factor for back*/
	int algo;		/* fan or parallel */
	float *fproj;		/* fan-beam projection pointers */
	float *pproj;		/* parallel projection, in gamma */
	float *proj;		/* parallel projection, in t */
	float *ggamma;		/* array of detector angles */
	float convert;		/* pi / 180 */
	MODE imode;		/* imode from input data */
	float slide;		/*z-slide [mm] per 360 rotation*/
	float cenp;		/* center detector - parallel */
	float cenf;		/* center detector - fan */
	float wdiso;		/* width of detector at iso-center */
	int iblk,nblki;		/* input block pointer and counter */
	int oblk,nblko;		/* output block pointer and counter */
	FILE *ifd,*ofd;		/* input/output file descriptors */
	float dgamma;		/* angle subtended by one detector */
	HU hu;			/* header for input and output files*/
	float delta;		/* fan angle */
	float pi;		/* ... in you face */
	float x;		/* distance on flat detector array */
	float dtheta;		/* parallel rotation angle delta */
	float theta0;		/* first parallel rotation angle */
	float theta;		/* first parallel rotation angle */
	int l,i,ii,nv,ns,ip;	/* loopers and pointers */
	float ir;		/* interpolation offsets */
	int d;			/* interpolation kernel */
	float t;		/* 't' location of parallel projection */
	float *tp;		/* pointers into fan projection based on t */
	float *aik;		/* azimuthal interpolation kernels */
	float *rik;		/* radial interpolation kernels */
	float *kp;		/* kernel pointer */
	int aoff,roff;		/* interpolation offsets */
	float pa;		/* projection accumulator */
/*
	parse command line
*/
	parse_all_basic(".rebin","REBIN",argc,argv);
/*
	open files
*/
	ifn = parse_suffix_malloc(ifn,"da");
	ofn = parse_suffix_malloc(ofn,"da");
	ifd = parse_fopen(ifn,"r");
	ofd = parse_fopen(ofn,"w");
/*
	get header information
*/
	ct_read_header(ifd,&hu);

	nf = hu.heads[CT_N];
	k = hu.heads[CT_K];
	algo = hu.heads[CT_ALGO];
	nrot = hu.heads[CT_ROT];		
	imode = hu.heads[CT_MODE];		

	r = hu.headf[CT_R];
	rsc = hu.headf[CT_RSC];
	beta0 = hu.headf[CT_ALPHA];
	dbeta = hu.headf[CT_DBETA];
	scale = hu.headf[CT_SCALE];
	slide = hu.headf[CT_SLIDE];
	cenf = hu.headf[CT_CEN];
	wdiso = hu.headf[CT_WDISO];
/*
	check parameter validity
*/
	if(algo == CT_PAR)parse_er("rebin: can't process parallel projections");
	parse_pcheck(imode,0,MAX_MODE,"rebin: illegal projection mode");
	parse_pcheck(aio,1,1024,"rebin: azimuthal interpolation order");
	parse_pcheck(rio,1,1024,"rebin: radial interpolation order");
/*
	set up geometry
*/
	pi = M_PI;
	convert = pi / 180.0;
	dtheta = dbeta;
	theta0 = beta0;
	delta = 2 * asin(r/rsc);
	dgamma = atan(wdiso/rsc);

	if(iso == 1){
		np = nf;
		cenp = cenf;
	}else{
		np = 2 * r / wdiso;
		cenp = floor(np/2) + cenf - floor(cenf);
	}
/*
	set up buffers
*/
	nblki = BLOCK_FLOAT(nf);
	fproj = parse_malloc_float(nf*k);
	ggamma = parse_malloc_float(nf);
	pproj = parse_malloc_float(nf);

	nblko = BLOCK_FLOAT(np);
	i = nblko * 512 / sizeof(float);
	tp = parse_malloc_float(np);
	proj = parse_malloc_float(i);

	aik = parse_malloc_float(intqa * aio);
	rik = parse_malloc_float(intqr * rio);
/*
	set up interpolation kernels
*/
	aoff = aio/2 - 1;
	build_itbl(aik,aio,aoff,intqa);

	roff = rio/2 - 1;
	build_itbl(rik,rio,roff,intqr);
/*
	set up algorithm independent information
*/
	if(algo == CT_SPACE){
		for(i=0;i<nf;i++){
			x = (i-cenf) * wdiso;
			ggamma[i] = atan2(x,(float)rsc);
		}
	}else if(algo == CT_ANGLE){
		for(i=0;i<nf;i++){
			ggamma[i] = (i-cenf) * dgamma;
		}
	}

	for(i=0;i<np;i++){
		t = (i - cenp) * wdiso;
		if(algo == CT_ANGLE){
			tp[i] = asin(t/rsc) / dgamma + cenf;
		}else{
			tp[i] = sqrt(t*t*rsc*rsc/(rsc*rsc-t*t));
			if(t < 0) tp[i] = -tp[i];
			tp[i] = tp[i] / wdiso + cenf;
		}
	}
	if(debug){
		parse_wda(tp,np,1,"tp");
	}
/*
	print program status
*/
        if(print){
		parse_wtext("******************* rebin *******************");
                parse_pdt("---start time");
		parse_rcsid_print(rcsid);
                parse_prs("input file:",ifn);
                parse_prs("output file:",ofn);
                parse_pri("n (fan) =",nf);
                parse_pri("n (parallel) =",np);
                parse_pri("k (total) =",k);
                parse_pri("k (rot) =",k/nrot);
                parse_pri("nrot =",nrot);
		parse_prf("rsc =",rsc);
		parse_prf("r =",r);
		parse_prf("dbeta =",dbeta);
		parse_prf("fan angle (degrees) =",delta/convert);
		parse_prf("center detector (fan) =",cenf);
		parse_prf("center detector (parallel) =",cenp);
                parse_prf("wd (iso) =",wdiso);
		parse_prf("slide (mm) =",slide);
		parse_pri("aio =",aio);
		parse_pri("rio =",rio);
		parse_pri("aq =",intqa);
		parse_pri("rq =",intqr);
		if(iso == 0)parse_wtext("isocenter interpolation: normal");
		if(iso == 1)parse_wtext("isocenter interpolation: none");
		if(iso == 2)parse_wtext("isocenter interpolation: only");

		ct_print_algo(algo);
		ct_print_mode(imode);
        }
/*
	write header information
*/
        parse_zbuf(hu.heads,256);

	hu.heads[CT_N] = np;
	hu.heads[CT_K] = k;
	if(iso == 1){
		hu.heads[CT_ALGO] = algo;
	}else{
		hu.heads[CT_ALGO] = CT_PAR;
	}
	hu.heads[CT_ROT] = nrot;		
	hu.heads[CT_MODE] = imode;		

	hu.headf[CT_R] = r;
	hu.headf[CT_RSC] = rsc;
	hu.headf[CT_ALPHA] = theta0;
	hu.headf[CT_DBETA] = dtheta;
	hu.headf[CT_SCALE] = scale;
	hu.headf[CT_SLIDE] = slide;
	hu.headf[CT_CEN] = cenp;
	hu.headf[CT_WDISO] = wdiso;

        ct_write_header(ofd,&hu);

	oblk = 1;
/*
	read all the fan projections
*/
	iblk = 1;
	for(ii=0;ii<k;ii++){
		parse_fread_block((char *)(fproj+ii*nf),sizeof(float),nf,ifd,iblk);
		iblk += nblki;
	}
	parse_from_can_floatv(fproj,nf*k);
	parse_fclose(ifd);
/*
	loop through all the views
*/
	for(nv=0;nv<k;nv++){
		theta = theta0 + nv*dtheta;

				/* generate hybrid projection p(theta,gamma) */
		if(iso != 2){
			for(ns=0;ns<nf;ns++){
				ir = (theta - beta0 - ggamma[ns]) / dbeta;
				i = floor(ir);
				d = (ir - i) * intqa;
				kp = aik + d*aio;
				for(pa=l=0;l<aio;l++){
					ip = smod(l+i-aoff);
					pa += fproj[ip*nf+ns] * *kp++;
				}
				pproj[ns] = pa;

				if(nv == 0 && debug == 2){
					printf("nv,ns,ir,i,d %d %d %f %d %d\n",nv,ns,ir,i,d);
				}
			}
		}else{
			for(ns=0;ns<nf;ns++){
				pproj[ns] = fproj[nv*nf+ns]; 
			}
		}

				/* generate parallel projection p(theta,t) */
		if(iso != 1){
			for(ns=0;ns<np;ns++){
				ir = tp[ns];
				i = floor(ir);
				d = (ir - i) * intqr;
				kp = rik + d*rio;

				if(nv == 0 && debug == 2){
					printf("nv,ns,ir,i,d %d %d %f %d %d\n",nv,ns,ir,i,d);
				}

				i -= roff;
				for(pa=l=0;l<rio;l++){
					if(i<0){
						ip = 0;
					}else if(i >= nf){
						ip = nf - 1;
					}else{
						ip = i;
					}
					pa += pproj[ip] * *kp++;
					i++;
				}
				proj[ns] = pa;
			}
		}else{
			for(ns=0;ns<np;ns++){
				proj[ns] = pproj[ns];
			}
		}

		parse_to_can_floatv(proj,np);
		parse_wblock(ofd,oblk,(char *)proj,nblko);
		oblk += nblko;
	}
/*
	clean up and exit
*/
        if(print==1)parse_pdt("---stop time");
        exit(0);
}
